from typing import Any, Generator, List
from abc import *
from thonnycontrib.exceptions import NoFunctionSelectedToTestException
from .ast_parser import L1DocTest, L1TestAstParser
import ast

class TestFinderStrategy(ABC):
    """
    This is an implementation of the Strategy pattern. The strategies defines 
    the way that the ast nodes are extracted. 
    For example, the `FindAllStrategy` strategy will try to find all the ast nodes
    and extract them from the parsed source.
    
    All strategies rely on the AST module of python. The source is first analyzed 
    and all supported nodes are revealed, then each concrete strategy can find 
    its suitable nodes as needed. 
    """
    def __init__(self, filename="", source="", ast_parser:L1TestAstParser=None) -> None:
        self._ast_parser = L1TestAstParser(filename, source) if not ast_parser else ast_parser
    
    # ###### MAIN METHOD ###### #
    
    def find_l1doctests(self) -> Generator[L1DocTest, Any, None]:
        """Invoke this method to get the docstrings depending of the strategy.
        
        This method follows the Template Method pattern. It defines the common 
        algorithm for all concrete strategies. Then, it invokes the `_find_docstring()` 
        abstract method.

        Returns:
            Generator[L1DocTest, Any, None]: A generator of L1DocTest.
        """
        return self._find_l1doctests()
    
    @abstractmethod
    def _find_l1doctests(self) -> Generator[L1DocTest, Any, None]:
        pass

    def get_ast_nodes(self):
        return self._ast_nodes 
    
    def set_ast_nodes(self, ast_nodes):
        self._ast_nodes = ast_nodes
    
    def set_filename(self, filename):
        self._ast_parser.set_filename(filename)
    
    def set_source(self, source):
        self._ast_parser.set_source(source)
    
    
class FindAllL1DoctestsStrategy(TestFinderStrategy):
    def _find_l1doctests(self):
        return self._ast_parser.parse()


class FindSelectedL1DoctestStrategy(TestFinderStrategy):
    def __init__(self, selected_line) -> None:
        super().__init__()
        self._selected_line = selected_line
    
    def _find_l1doctests(self):
        """
        Note: The following docstring specifies only the raised excpetion.
        @See the docstring of the superclass to see the full docstring. 
            
        Raises:
            NoFunctionSelectedToTestException: When the selected line doesn' correspond to a 
            function or a class declaration.
        """
        selected_node = next(self._ast_parser.parse(self._selected_line), None)
        if not selected_node :
            msg = "%s\n\n%s" %  ("No function is selected to test !", 
                                 "The selected line must have a function or a class declaration.")
            raise NoFunctionSelectedToTestException(msg)
        yield selected_node
    
    def get_selected_line(self):
        return self._selected_line    
    
    def set_selected_line(self, selected_line):
        self._selected_line = selected_line   

class L1TestFinder:
    """
        The `TestFinder` relies on the `AST` module to parse the script and extracts 
        the nodes with their docstrings. 
        
        For each node, it invokes the `DoctestParser` to parse its docstring. Then, 
        for each dectected test, it associates an `Example` type that will contains 
        the `source`, `want` and the `line` of the test.
        
        Args:
            - filename(str): the filename of the source code.
            - source(str): the source code to be parsed by the `AST` parser and by the 
                    `DoctestParser`.
            - strategy(TestFinderStrategy): the strategy to use to find the docstrings.
    """
    def __init__(self, filename:str="", source:str="", strategy:TestFinderStrategy=None):      
        self._filename = filename
        self._source = source
        self._strategy = FindAllL1DoctestsStrategy(filename, source) if not strategy else strategy
    
    def find_l1doctests(self) -> List[L1DocTest]: 
        """
        Extract examples from the docstings of all the ast nodes of the source code.
        
        Returns a list of L1DocTest. Each of L1DocTest corresponds to an ast node.
        Each L1DocTest has it's list of `Example`
        
        Raises: 
            NoFunctionSelectedToTestException: when the selected line desn't refer to a 
            function/class signature.
            Error: A compilation error raised by the AST module.
            SpaceMissingAfterPromptException: when a space is missing after the prompt.
        """   
        generator = self._strategy.find_l1doctests()
        l1doctests: List[L1DocTest] = []
        while (l1doctest := next(generator, None)) is not None: 
            l1doctests.append(l1doctest)            
        return l1doctests
        
    def get_filename(self):
        return self._filename
    
    def set_filename(self, filename: str):
        self._filename = filename
        self._strategy.set_filename(filename)
    
    def get_source(self):
        return self._source
    
    def set_source(self, source: str):
        self._source = source
        self._strategy.set_source(source)
        
    def get_strategy(self):
        return self._strategy
    
    def set_strategy(self, strategy: TestFinderStrategy):
        self._strategy = strategy
        self._strategy.set_filename(self._filename)
        self._strategy.set_source(self._source)


from thonny.editors import EditorCodeViewText
import re
import tkinter as tk
from thonny import ui_utils
from thonnycontrib import utils

# The regex to check if the __main__ already exists
# The regex checks if a string already contains the __main__ and not commented.
_REGEX = re.compile(r"(?m)^\s*if\s+__name__\s*==\s*['\"]__main__['\"]\s*:\s*")

# The generated `__main__` function. 
# Leave a blank line before the `if __name__ == '__main__':` line, 
# so that the generated main function is separated from the rest of the code.
GENERATED_MAIN:str = (
"""
    
if __name__ == '__main__':
    # éxécuté qd ce module n'est pas initialisé par un import.
    pass
"""
)

# used to tell the user whether the __main_ is generated or not
MAIN_EXISTS_MSG = "`__main__` existe déjà"
MAIN_GENERATED_MSG = "`__main__` généré"

class MainGenerator(): 
    """
    This class is responsible for generating the __main__ function.
    It takes a source code as input and generates the __main__ function 
    at the bottom of the source code.
    
    Args:
        source (str, optional): The source code in which 
        the __main__ will be generated. Defaults to "".
    """         
    def __init__(self, source=""):
        self._source = source
     
    def generate(self, text_widget:EditorCodeViewText=None, show_tooltip_info=True) -> str: 
        """
        Generate the __main__ and insert it at the bottom of the given `text_widget`.

        Args:
            text_widget (EditorCodeViewText, optional): The text widget to insert 
            the generated __main__ into. Set to `None` to not insert the generated 
            main function, this is useful for testing.
            
            show_tooltip_info (bool, optional): Set to True to show a message on a tooltip 
            to tell the user whether the __main__ is generated or not. Defaults to True.
            
        Returns:
            str: The generated __main__ function or an empty string if the __main__ already exists.
        """        
        main_line = self.__find_main_lineno(self._source) # Check if a __main__ already exists
        if main_line:
            if show_tooltip_info:
                self.__show_info_on_tooltip(text_widget, main_line=main_line, msg=MAIN_EXISTS_MSG, is_success=False)
            return "" # __main__ already exists so return empty string.

        generated_main = self.__generate_main()
        if text_widget :  
            text_widget.insert("end", generated_main)
            if show_tooltip_info:
                last_line_number = int(text_widget.index('end').split('.')[0]) - generated_main.count("\n")
                self.__show_info_on_tooltip(text_widget, main_line=last_line_number, msg=MAIN_GENERATED_MSG, is_success=True)
        return generated_main
    
    def __find_main_lineno(self, text: str) -> int:
        """
        Returns the line numbers where the __main__ (and not commented) exists. Returns None if not found.
        """
        lines = text.splitlines()
        # use next() to get the first line number where the regex is found. Gets None if not found
        return next((i for i, line in enumerate(lines, 1) if re.search(_REGEX, line)), None)
    
    def __generate_main(self) -> str:
        """ Generate the main function """
        return GENERATED_MAIN
      
    def __show_info_on_tooltip(self, text_widget:EditorCodeViewText, main_line:int, msg:str, is_success:bool):
        """
        Show a message on a tooltip. A tooltip is a toplevel window. this method 
        customizes the style of the tooltip and it is destroyed after few seconds.

        Args:
            text_widget (EditorCodeViewText): The text widget where the tooltip will be displayed.
            main_line (int): The line where the existed __main__ is.
            msg (str): The message to display on the tooltip.
            is_success (bool): Set to True to set the colors of the tooltip to green, When set to False, 
            the colors are red.
        """
        wx, wy, _, _ = text_widget.bbox("insert")
        lineno, column = utils.get_selected_line(text_widget, only_lineno=False)
        
        tw = tk.Toplevel(text_widget)   
        tw.after(1500, tw.destroy) # destroy the tooltip after 1.5 seconds
        tw.wm_overrideredirect(1) # remove the title bar
        tw.wm_geometry("+%d+%d" % (column+wx+text_widget.winfo_rootx(), lineno+wy+text_widget.winfo_rooty()))

        options = ui_utils.get_style_configuration("Tooltip").copy()
        
        # (#facdcd, #850303) are the default colors for error messages
        # (#d4f4cd, #0b4f00) are the default colors for success messages
        options.setdefault("background", "#d4f4cd" if is_success else "#facdcd")
        options.setdefault("foreground", "#0b4f00" if is_success else "#850303")
        options.setdefault("relief", "flat") # remove the border
        options.setdefault("padx", 1)
        options.setdefault("pady", 0)

        msg = " %s, ligne %s " % (msg, main_line)
        label = tk.Label(tw, text=msg, **options)
        label.pack()

    def get_source(self) -> str:
        """ Returns the source code """
        return self._source
    
    def set_source(self, source: str):
        """ Sets the source code """
        self._source = source


import unittest
from thonnycontrib. main_generator.main_generator import MainGenerator, GENERATED_MAIN

class TestMainGenerator(unittest.TestCase):
    def setUp(self):
        self.main_generator = MainGenerator(source="")
        
    def tearDown(self):
        self.main_generator = None

    def test_generate_with_no_existing_main(self):
        generated: str = self.main_generator.generate(show_tooltip_info=False)
        self.assertTrue(bool(generated)) # generated is not empty
        self.assertEqual(GENERATED_MAIN, generated)

    def test_generate_with_existing_main(self):
        self.main_generator.set_source("if __name__ == '__main__':")
        generated = self.main_generator.generate(show_tooltip_info=False)
        self.assertEqual(generated, "") # ensure nothing is generated
        
    def test_generate_with_existing_commented_main(self):
        self.main_generator.set_source("# if __name__ == '__main__':")
        generated = self.main_generator.generate(show_tooltip_info=False)
        self.assertTrue(bool(generated)) # ensure that __main__ is generated
        self.assertEqual(GENERATED_MAIN, generated)

    def test_find_main_lineno(self):
        text = (
                """\
                print('hello, world')
                if __name__ == '__main__':
                    print('this script is being run directly')
                """
            )
        lineno = self.main_generator._MainGenerator__find_main_lineno(text) # this is how we call private methods in python just to test them
        self.assertEqual(lineno, 2)
        
    def test_find_main_lineno_when_commented(self):
        text = (
                """\
                print('hello, world')
                # if __name__ == '__main__':
                #    print('this script is being run directly')
                """
            )
        lineno = self.main_generator._MainGenerator__find_main_lineno(text)
        self.assertEqual(lineno, None)

if __name__ == "__main__":
    unittest.main(verbosity=2)



from thonny.config_ui import ConfigurationPage
from ..properties import PLUGIN_NAME
from thonny import get_workbench

# Default config
DEFAULT_DOC_GENERATION_AFTER_RETURN = True
DEFAULT_IMPORT_MODULE_IN_SHELL = True
DEFAULT_CLOSE_FUNCTION_ROWS = False
DEFAULT_OPEN_RED_TESTS = False
DEFAULT_OPEN_ONLY_RED_FUNCTIONS = True
DEFAULT_HIGHLIGHT_EXCEPTIONS = False
DEFAULT_REPORT_EXCEPTION_DETAIL = True 

# Option names
AUTO_GENERATON_DOC = "auto_generaton_doc"
IMPORT_MODULE = "import_module"
OPEN_RED_TEST_ROWS = "open_red_rows"
CLOSE_FUNCTION_ROWS = "close_function_rows"
OPEN_ONLY_RED_FUNCTIONS = "open_only_red_functions"
HIGHLIGHT_EXCEPTIONS = "highlight_exceptions"
REPORT_EXCEPTION_DETAIL = "exception_detail"

# Dict of options name and default value
OPTIONS = {
    AUTO_GENERATON_DOC : DEFAULT_DOC_GENERATION_AFTER_RETURN,
    IMPORT_MODULE : DEFAULT_IMPORT_MODULE_IN_SHELL,
    CLOSE_FUNCTION_ROWS: DEFAULT_CLOSE_FUNCTION_ROWS,
    OPEN_RED_TEST_ROWS : DEFAULT_OPEN_RED_TESTS,
    OPEN_ONLY_RED_FUNCTIONS : DEFAULT_OPEN_ONLY_RED_FUNCTIONS,
    HIGHLIGHT_EXCEPTIONS: DEFAULT_HIGHLIGHT_EXCEPTIONS,
    REPORT_EXCEPTION_DETAIL: DEFAULT_REPORT_EXCEPTION_DETAIL
}

def init_options():
    """
    Initialise dans le workbench les options du plugin.
    """
    for opt in OPTIONS :
        if not get_workbench().get_option(opt) :
            get_workbench().set_default("%s." % PLUGIN_NAME + opt, OPTIONS[opt])

def get_option(name: str): 
    """
    Renvoie la valeur dans le workbench de l'option passée en paramètre.

    Paramètres:
    - name : le nom de l'option, tel que définie ds globals.py 
    """
    return get_workbench().get_option("%s." % PLUGIN_NAME + name)

def set_option(name, value):
    get_workbench().set_option("%s." % PLUGIN_NAME + name, value)

class L1TestConfigurationPage(ConfigurationPage):
    def __init__(self, master):
        ConfigurationPage.__init__(self, master)
        
        self.add_checkbox("%s.%s" % (PLUGIN_NAME, AUTO_GENERATON_DOC), 
                          "Générer la docstring automatiquement après un saut de ligne au niveau du nom d'une fonction.")

        self.add_checkbox("%s.%s" % (PLUGIN_NAME, IMPORT_MODULE), 
                          "Importer le module exécuté dans le shell.")
        
        self.add_checkbox("%s.%s" % (PLUGIN_NAME, CLOSE_FUNCTION_ROWS),
                          "Fermer toutes les lignes de fonctions dans la vue %s." % PLUGIN_NAME)
         
        self.add_checkbox("%s.%s" % (PLUGIN_NAME, OPEN_ONLY_RED_FUNCTIONS),
                          "Ouvrir seulement les lignes de fonctions rouges dans la vue %s." % PLUGIN_NAME)
        
        self.add_checkbox("%s.%s" % (PLUGIN_NAME, OPEN_RED_TEST_ROWS), 
                          "Ouvrir les tests en echec dans la vue %s." % PLUGIN_NAME)
        
        self.add_checkbox("%s.%s" % (PLUGIN_NAME, HIGHLIGHT_EXCEPTIONS), 
                          "Mettre en évidence les tests en échec (seulement ceux qui lèvent une exception).")



# Module Evaluator

from typing import List

from thonnycontrib.utils import send_current_state
from .ast_parser import L1DocTest
from thonnycontrib.properties import FINISHED_STATE
from thonnycontrib.exceptions import *
from thonnycontrib.utils import *
from .doctest_parser import *
from thonny import *
from types import ModuleType
from .test_finder import L1TestFinder, TestFinderStrategy
import sys 

class Evaluator:
    """
    Instantiate an `Evaluator` with a filename, a source and a test_finder. 
    The evaluator relies on the `TestFinder` to parse and extracts the tests from
    the docstrings. 
    
    Args:
        filename (str, optional): the filename to be executed by the `Evaluator`. Defaults to "".
        source (str, optional): the source code to be parsed by the `TestFinder`. Defaults to "".
        test_finder (TestFinder, optional): the finder of the tests declared in docstrings. 
                                        Defaults to TestFinder().
    """
    def __init__(self, filename:str="", source:str="", test_finder=L1TestFinder()):
        self._globals = dict()
        self._module: ModuleType = None
        self._test_finder: L1TestFinder = test_finder
        self._test_finder.set_filename(filename)
        self._test_finder.set_source(source)

    def evaluate(self) -> List[L1DocTest]:
        """Imports the module from the filename, sets the global variables to
        imported module's dict and then evaluates the tests.

        Returns:
            List[L1DocTest]: Returns the list of the evaluated l1doctests.
        """
        # si le `self._module` est null, alors `Evaluator` va s'occuper d'importer lui même
        # le module à partir du filename. 
        # Sinon, le module a été fournit, donc `Evaluator` utilisera le module fournit.
        if not self._module:
            # The `import_module()` function can raise an exception(see it's doc)
            self._module = import_module(self.get_filename()) 
        
        # set a globals dictionary that will contains all the functions and 
        # meta-informations decalared in the module
        self.set_global_variables_to_module_ones()
        return self.__evaluate_l1doctests() 
    
    def __evaluate_l1doctests(self) -> List[L1DocTest]:
        """
        Evaluate all the extracted l1doctests containing in the source code.
        
        Args:
            module (ModuleType): The module to be run.
        
        Returns:
            List[L1DocTest]: Returns the list of the evaluated l1doctests. 
            The verdicts are already computed and contained in each of the L1doctest.
        
        Raises:
            NoTestFoundException: When no test are found in the working editor.
            CompilationError: When the doctest parser raises an exception, it's catched 
                            and raised as a `CompilationError`, or when AST parser fails.
        """
        try:     
            # This line parses the source using the AST module and can raise a compilation error 
            l1doctests = self._test_finder.find_l1doctests() 
            for l1doctest in l1doctests: 
                self.set_global_variables_to_module_ones()
                l1doctest.evaluate(self.get_globals())
            send_current_state(state=FINISHED_STATE) # tells that the evaluation is finished
            return l1doctests
        except NoFunctionSelectedToTestException as e:
            raise RuntimeException(str(e))
        except DoctestParserException as e:
            raise CompilationError(str(e)) 
        except BaseException as e:
            # the compilation error is catched and raised as a CompilationError
            # and the evaluation is interrupted(because we cannot parse a content
            # with compilation errors).
            error_info = sys.exc_info()
            formatted_error = get_last_exception(error_info)
            raise CompilationError(formatted_error)      
    
    def get_globals(self):
        return self._globals

    def get_module(self):
        return self._module

    def set_module(self, module: ModuleType):
        self._module = module
        
    def set_global_variables_to_module_ones(self):
        '''Set self._globals to module __dict__ using a shallow copy like
        doctest does.
        '''
        self._globals = self._module.__dict__.copy()
        
    def set_test_finder(self, test_finder):
        self._test_finder = test_finder
    
    def get_test_finder(self):
        return self._test_finder

    def set_filename(self, filename):
        self._test_finder.set_filename(filename)
    
    def get_filename(self):
        return self._test_finder.get_filename()
    
    def set_source(self, source:str):
        self._test_finder.set_source(source)
        
    def get_source(self):
        return self._test_finder.get_source()
    
    def set_finder_strategy(self, finder_strategy: TestFinderStrategy):
        self._test_finder.set_strategy(finder_strategy)





from thonny import get_workbench
from thonny.ui_utils import select_sequence
from .docstring_generator.doc_generator import DocGenerator
from .l1test_frontend.l1test_reporter import L1TestErrorView, L1TestTreeView
from .main_generator.main_generator import MainGenerator
from .exceptions import *
from .properties import  ERROR_VIEW_LABEL, PLUGIN_NAME
from .utils import (
    assert_one_line_is_selected,
    get_focused_writable_text,
    get_selected_line, 
    get_image_path
)
from .l1test_configuration.l1test_options import *
from .l1test_frontend import get_l1test_runner
from .l1test_frontend import get_outliner
from .environement_vars import *
from functools import partial
from .docstring_generator import get_doc_generator

def run_all_tests():
    """
    Cette fonction est invoquée quand le button `l1test` est cliqué.
    Cette fonction permet d'envoyer au l1test_backend la commande L1test.
    """
    get_l1test_runner().run_l1test()
      
def generate_docstring(auto: bool, event=None): 
    """
    Cette fonction est invoquée quand le button `doc_generator` est cliqué. Ou quand on 
    réalise un saut de ligne après la déclaration d'une fonction.

    Args:
        auto (bool):  Si True, alors la docstring sera générée automatiquement après un saut de ligne. 
        Sinon, la docstring sera générée quand le button `doc_generator` est cliqué.
        event (Any, optional): Si event est différent de None, alors un saut de ligne a été réalisé. 
        et donc la docstring sera générée automatiquement.
        Defaults to None.
    """
    if _writable_text_is_focused(): # on vérifie si la zone séléctionnée est une zone de l'éditeur
        text_widget = get_focused_writable_text()
        selected_line = get_selected_line(text_widget) 
        docGenerator = get_doc_generator(auto)  
        docGenerator.run(selected_line - 1 if auto else selected_line, text_widget)

def generate_main():
    """
    Generate the __main__ function.
    """
    text_widget = get_focused_writable_text()
    main_generator = MainGenerator(text_widget.get("1.0", "end"))
    main_generator.generate(text_widget)
    del main_generator # destruction de l'objet en mémoire (for better performance)
    

def _writable_text_is_focused():
    """
    Returns:
        boolean: Returns True if the selected zone is a writable text.
    """
    return get_focused_writable_text() is not None


def __init_l1test_options():
    init_options()
    get_workbench().add_configuration_page(PLUGIN_NAME, PLUGIN_NAME, L1TestConfigurationPage, 30)

def __init_l1test_views():
    get_workbench().add_view(L1TestTreeView, PLUGIN_NAME, "nw", visible_by_default=True)
    get_workbench().add_view(L1TestErrorView, ERROR_VIEW_LABEL, "sw", visible_by_default=False)

def __init_l1test_commands():
    # Création du button l1test au niveau de la barre des commandes
    get_workbench().add_command(command_id=PLUGIN_NAME,
                                menu_name=PLUGIN_NAME,  
                                command_label="Run all tests",
                                handler=run_all_tests,
                                include_in_toolbar=True, #j'inclue ici ce bouton dans la toolbar 
                                image=get_image_path("l1test_icon.png"),
                                caption=PLUGIN_NAME)
    
    # Création du button L1Test dans la barre de menu en haut.  
    get_workbench().add_command(command_id="Run one test",
                                menu_name=PLUGIN_NAME,  
                                command_label="Run tests for ...",
                                image=get_image_path("l1test_icon.png"),
                                submenu=get_outliner().get_menu()
    )
    
    # Création du bouton dans le menu 'Edit' pour lancer la génération de docstring
    get_workbench().add_command(command_id="doc_generator",
                                menu_name="edit",  
                                command_label="Generate a docstring",
                                handler=partial(generate_docstring, auto=False), 
                                tester=_writable_text_is_focused and assert_one_line_is_selected,
    )
     
    # Création du bouton dans le menu 'Edit' pour générer le main
    get_workbench().add_command(command_id="main_generator",
                                menu_name="edit",  
                                command_label="Generate a main",
                                handler=generate_main, 
                                tester=_writable_text_is_focused,
                                default_sequence=select_sequence("<Alt-m>", "<Command-Alt-m>", "<Alt-m>"),
                                accelerator="Alt+m"
    )

def __add_event_binding():
    # Quand un saut de ligne est réalisé après la déclaration d'une fonction,
    # alors une docstring sera générée automatiquement.
    get_workbench().bind_class("EditorCodeViewText", 
                               "<KeyRelease-Return>", 
                               lambda event: generate_docstring(auto=get_option(AUTO_GENERATON_DOC), event=event))

def load_plugin():
    """
    load_plugin est un nom de fonction spécifique qui permet à thonny de charger les élements du plugin
    """
    # these functions should be called in this order
    __init_l1test_options()
    __init_l1test_views()    
    __init_l1test_commands()
    __add_event_binding()



# Auteur : Réda ID-TALEB

from .FailedVerdict import FailedVerdict

class FailedWhenExceptionExpectedVerdict(FailedVerdict):
    def __init__(self, filename:str, lineno, tested_line, expected_result, failure_message=""):
        super().__init__(filename, lineno, tested_line, expected_result, failure_message)

    def isSuccess(self):
        return False
    
    def get_details(self):
        common = "%s was not raised by `%s`" % (self.expected_result, self.tested_line)
        if self.details:
            common += "\n%s" % self.details
        return common

    def _verdict_message(self):
        return super()._verdict_message()


# Auteur : Esteban COLLARD, Nordine EL AMMARI
# Refactored by Manal LAGHMICH & Réda ID-TALEB
# Last refactored by Réda ID-TALEB

from abc import ABC, abstractmethod

class ExampleVerdict(ABC):
    def __init__(self, filename:str, lineno:int, tested_line:str, expected_result:str, details:str=""):
        """
        Build an instance of a verdict. A the Test object represents a result of a 
        tested line in a docstring.
        
        Args:
            filename (str): The filename in wich the tests are run.
            lineno (int): The line number of the tested line.
            tested_line (str): The tested line.
            expedted_result (str): The expected result of the tested line.
            details (str): Other details that may be added to the verdict.
        """
        self.filename = filename
        self.lineno = lineno
        self.tested_line = tested_line.strip("\n").strip()
        self.expected_result = expected_result.strip("\n").strip()
        self.details = details
    
    @abstractmethod
    def get_color(self) -> str:
        """The color of the verdict."""
        pass
    
    @abstractmethod
    def isSuccess(self) -> bool:
        """Returns True if the verdict is a success, False otherwise."""
        pass
    
    @abstractmethod
    def _verdict_message(self) -> str:
        """
        Return the verdict message. But should use the __str__ method 
        instead by calling str(verdict)
        """
        pass
    
    def __str__(self) -> str:
        return "%s for `%s`" % (self._verdict_message(), self.tested_line)
    
    @staticmethod
    def get_verdicts_by_priority():
        from .ExceptionVerdict import ExceptionVerdict
        from .PassedVerdict import PassedVerdict
        from .FailedVerdict import FailedVerdict
        from .FailedWhenExceptionExpectedVerdict import FailedWhenExceptionExpectedVerdict
        return [ExceptionVerdict,
                FailedWhenExceptionExpectedVerdict, 
                FailedVerdict, 
                PassedVerdict]
    
    ## Getters ##
    def get_tested_line(self):
        return self.tested_line

    def get_expected_result(self):
        return self.expected_result
    
    def get_details(self):
        return self.details
    
    def get_filename(self):
        return self.filename
    
    def get_lineno(self):
        return self.lineno


from collections import namedtuple
from typing import List

from ..utils import clear_env_vars
from ..backend.ast_parser import L1DocTest
from ..exceptions import *
from ..environement_vars import *
from .l1test_reporter import *
from ..properties import *
from thonny.common import ToplevelResponse, InlineResponse
from thonny.workbench import WorkbenchEvent
from thonnycontrib import l1test_frontend 
from thonny.editors import EditorNotebook
from thonny import editors
import thonny
from ..ThonnyLogsGenerator import log_in_thonny
import pickle, inspect, thonny, time
from ..l1test_configuration import l1test_options 


# ErrorMsg déclare deux champs : le préfixe et le message d'erreur
# le préfix est juste le titre qui précède le message d'erreur sur la Error view
ErrorMsg = namedtuple("ErrorMsg", "title msg")

# Le nom de l'event qui lance le redémarrage du backend thonny
BACKEND_RESTART_EVENT = "BackendRestart"

class L1TestRunner:
    """
        The L1TestRunner is responsible for receiving the test results returned 
        by the l1test_backend and displaying them on the treeview. The L1TestRunner 
        is also responsible for displaying errors if the l1test_backend returns 
        a response that carries an exception.
        
        Finally, the L1TestRunner decides which view (the Treeview or the ErrorView)
        to display depending on the state of program execution.
        
        Note: L1TestRunner does not deal with the construction of the views but 
        it allows to invoke the correct one.
    """
    def __init__(self, reporter=None):
        l1test_frontend._l1test_runner = self
        
        self._reporter = L1TestReporter() if not reporter else reporter
        self._has_exception = False
        self._is_l1test_running = False
        self._is_pending = False
        self._has_reported_exception = False # indique si une ligne d'erreur de la tree a été clique et que l'exception
                                            # a été reporté sur la ErrorView
        self._filename = "<module>"
        # Quand le backend envoie une réponse de type `ToplevelResponse``,
        # alors le TestRunner va invoquer la fonction `show_verdicts`
        thonny.get_workbench().bind("ToplevelResponse", self._handle_backend_response, True)
        
        # Quand le backend envoie une réponse de type `InlineResponse``,
        # alors le TestRunner va invoquer la fonction `show_execution_state`
        thonny.get_workbench().bind("InlineResponse", self._handle_execution_state, True)
        
        thonny.get_workbench().bind(L1TREE_VIEW_EVENT, self._handle_clicked_exception, True)
        
        # Quand le backend est redémarré en thonny nous invoquons la méthode 
        # `self._on_restart_backend()`
        thonny.get_workbench().bind(BACKEND_RESTART_EVENT, self._on_restart_backend, True)
    
    def run_l1test(self, selected_line: int=None):
        """
        Run the L1Test plugin by sending a command to the l1test_backend. This method checks
        if the file is saved and if the editor is empty. If the file is not saved or the editor
        then it will display an error message on the ErrorView. Otherwise, it will send a command 
        to the backend to run the tests.
        Args:
            is_selected (bool): Set as True if only one method is selected to run
                            it's tests.
            selected_line (int): The number of the selected line. 
            Default is 0 if no line is selected.
        """
        try:
            editor: EditorNotebook = get_workbench().get_editor_notebook()
            # si aucun editeur n'est ouvert sur le workbench
            if not editor.get_current_editor():
                raise NoEditorFoundException("No editor found !\n\nPlease open an editor before running the tests.")
            
            # cette ligne demande de sauver le fichier s'il n'a pas encore été sauvé sur
            # la machine. Si le fichier est déjà sauvé, il va permettre d'enregistrer la nouvelle
            # version du fichier.
            filename = editors.get_saved_current_script_filename(force=True)          

            # si le filename est null alors le fichier n'a  pas été sauvé sur machine.  
            # Ce cas survient quand l'utilisatur quitte la fenetre de sauvegarde sans sauver le fichier.
            if not filename: 
                msg = "The file is not saved.\n\nConsider to save the file before running the tests."
                raise NotSavedFileException(msg)

            self._filename = filename
            
            if not editor.get_current_editor_content().strip():  # L'éditeur est vide. 
                # on a pas envie d'envoyer une commande au backend si le fichier est vide.
                # Dans tous les cas y a rien à tester.
                raise EmptyEditorException("The editor is empty!\n")
            
            # si on est là alors le fichier est bien sauvegardé et contient quelque chose.
            self.request_backend(selected_line)         
        except FrontendException as e: # on catche que les exception coté view
            self.terminate_running()
            self.set_has_exception(True) 
            self.show_error_view(str(e))
    
    def request_backend(self, selected_line:int):
        """Allows to execute the `L1test` magic command. 
        
        There's two cases : 
        1. if the `L1test` is invoked only one time so the command is sent to 
        backend to be executed by the thonny's runner. 
        2. if the `L1test` is invoked a lot of times (lot of clicks), then we only
        consider the first invocation (first click) of the `L1test` command. While 
        the l1test still running, the other clicks are ignored.

        Args:
            selected_line (int): The number of the selected line.
        """
        if not self.is_running(): # si l1test n'est est pas déjà en cours d'execution
            try:
                self.set_is_running()
                self.__request_backend(selected_line=selected_line)
            except:
                self.set_is_running(False)
                treeview = self._reporter.get_treeview()
                treeview.insert_in_header("Coudn't restart the backend.\nPlease restart with the 'Stop' button !", 
                                        clear=True, tags=("red"), image=ERROR_ICON)
                thonny.get_runner().restart_backend(clean=True)
                time.sleep(1)
                self.request_backend(selected_line=selected_line)
            finally:
                clear_env_vars(IMPORT_MODULE_VAR, SELECTED_LINE_VAR)
        
    def __request_backend(self, command_name=BACKEND_COMMAND, selected_line:int=None):
        """
        Sends a command to the Thonny's backend to execute the current script.
        The name of the command must starts with a uppercase.    
        
        Please note that `thonny.get_runner.execute_current()` will consider 
        all the current script, and it builds a `ToplevelCommand` automatically. 
        So in the case if you don't want to consider all the script but only a 
        selected method, you should set the number of the selected line. 
        
        The value of the selected line is stored in an environnement variable to be
        shared with the backend. The handler of the command(l1test_backend) can access 
        this environnement variables to decide if the `Evaluator` should run all the 
        tests or only for the selected line.

        Note: the value of the environnement variable should be a string.
        For example : 
        ```py
        os.environ["is_selected"] = str(is_selected) # the value should always be a string 
        ```
        
        In the handler of the command :
        ```py
        def _cmd_l1test(cmd: ToplevelCommand):         
            is_selected: bool = eval(os.environ.get("is_selected"))
        ```
                    
        Note : when a new command is called in Thonny it triggers a partial 
        "restart" of the backend before processing the new command. Thonny does 
        this to force stop the current process and start a new process for the 
        new command.
            
        Args:
            command_name (str): A command name , should starts with a upper case. 
            Defaults to BACKEND_COMMAND.
            selected_line (int, optional): The number of the selected line. Defaults to None
            if no line is selected.
        """ 
        # on stocke dans une variable d'environnment la valeur de l'option `import_module in shell`.
        # le l1test backend retrouvera cette valeur pour décider d'importer ou non le module dans le shell.
        os.environ[IMPORT_MODULE_VAR] = str(l1test_options.get_option(l1test_options.IMPORT_MODULE))
        
        if selected_line:
            os.environ[SELECTED_LINE_VAR] = str(selected_line)
        
        # Une fois que execute_current() est executée un restart backend partiel est invoqué
        # pour un nouveau processus pour la commande passé en paramètre.
        thonny.get_runner().execute_current(command_name)

    def _handle_backend_response(self, msg: ToplevelResponse):     
        """
        This method is binded to the `TopLevelResponse` event sent by the backend.
        
        If this method is triggered so the `msg` parameter will contain the response received 
        from the backend. Please note that the `TopLevelResponse` event is not necessary sent 
        by the l1test_backend, but it can be also sent by the Shell. The shell contains an 
        internal infinite loop that waits for commands and sends the responses periodically.
        
        This method verify the source of the `TopLevelResponse` event. If the source is 
        l1test_backend so it checks the recieved response if it contains exception or not. 
        If the response contains an exception so the according error will be displayed in 
        error view. Otherwise the response contains the verdicts of the tests and will be shown 
        on the treeview.
        
        Note: The data is deserialized before displaying it on the view.
        """   
       
        # On vérifie si le TopLevelRespone reçu est envoyé par le l1test_backend 
        if self._is_relevant_response(msg):
            verdicts, error_msg = self.__get_verdicts_and_error(msg)
            if not self.has_exception(): 
                self.set_and_show_verdicts_on_treeview(verdicts) 
            else: 
                self.clean_treeview()
                self.show_error_view(error_msg.msg, error_msg.title)
        else:
            # Le TopLevelReponse reçu ne nous intéresse pas.
            return 
        # On indique l'état de l'execution du la commande comme terminée
        self.terminate_running()
    
    def __get_verdicts_and_error(self, msg:ToplevelResponse) -> tuple[List[L1DocTest], ErrorMsg]:
        """
        Returns the l1doctests and the error message from the received response. 
        If the response contains an exception so the error message will be returned. In this
        case the l1doctests will be None. Reverse is true, if the response does not contain the
        error message so the l1doctests will be returned and the error message will be None.
        
        Args:
            msg (ToplevelResponse): The received response from the backend.

        Returns:
            a tuple of (l1doctests, error_msg)
        """
        l1doctests, error_msg = None, ErrorMsg(title=None, msg=None)
        self.clean_error_view()
        if msg.get(L1TEST_EXCEPTION):
            error_msg = self.__handle_raised_exception(msg.get(L1TEST_EXCEPTION))
        if not self.is_pending() and self.is_running(): # si le plugin est en cours d'execution et n'est pas en attente (càd pas de tests en cours d'execution)
            self.clean_treeview()
            received_verdicts = msg.get(VERDICTS)
            if received_verdicts:
                self.set_has_exception(False)
                # deserialize the verdicts
                l1doctests: List[L1DocTest] = pickle.loads(received_verdicts)
                # log in thonny                    
                log_in_thonny(l1doctests, self._filename, os.environ.get(SELECTED_LINE_VAR) != None)
        return l1doctests, error_msg
    
    def __handle_raised_exception(self, exception: dict) -> ErrorMsg:
        """
        This function handles the raised exception by returning a tuple containing
        a prefix message and the exception message.

        The prefix refers to the title shown on the ErrorView. If you want to remove 
        the title, you should specify the prefix as an empty string. If you don't 
        want to change the title, you should specify the prefix as `None`.
        
        Note: this function also sets the state of the execution. It set the `_has_exception`
        attribute to True if the exception is not an `InterruptedError`. Otherwise, it
        sets the `_has_exception` attribute to False.
        
        Args:
            exception (dict): The exception to handle.

        Returns:
            ErrorMsg: a couple that indicates a prefix message and the exception message.
        """
        if exception.get("type_name") == InterruptedError.__name__:
            self.set_has_exception(False)  # ce n'est pas une erreur
            self.clean_treeview()
            
            return ErrorMsg(title=None, msg=None)    # On fait rien.
        
        self.set_has_exception(True)   # on indique l'état de l'execution
        
        # The backend exceptions inherit from the BackendException class
        backend_exceptions = self.__get_all_backend_exceptions()
        
        for backend_exception in backend_exceptions:                
            if (exception.get("type_name") == backend_exception.__name__) :
                return ErrorMsg(exception.get("prefix"), exception.get("message")) 
        
        # si on est là alors l'exception levée n'est pas une exception backend est
        # c'est probablement une exception levée par python.
        return ErrorMsg(title=exception.get("prefix"), 
                        msg="%s:\n%s" % (exception.get("type_name"), exception.get("message")))
    
    def __get_all_backend_exceptions(self) -> list[BackendException]:
        """
        All the exceptions (specifc to l1test) raised on the backend 
        inherit from the `BackendException` class.
        
        Gets dynamically the exceptions inherit from the `BackendException` class that
        are declared in the `exceptions.py` module.

        Returns:
            list[BackendException]: Returns a list of the backend exception classes.
        """
        import thonnycontrib.exceptions as exceptions
        backend_exceptions = []
        for name, obj in inspect.getmembers(exceptions):
            if inspect.isclass(obj) and issubclass(obj, exceptions.BackendException):
                backend_exceptions.append(obj)
        return backend_exceptions   
    
    def _on_restart_backend(self, event: WorkbenchEvent):
        """
        This function is called when the backend of thonny is restared. The restarting
        of the backend generates a `BackendRestart` event and this event can 
        be generated either by  the red `Stop/Restart backend` button in Thonny's 
        toolbar or by invoking a new command. 
        
        When a new command is called in Thonny it triggers a partial restart
        of the backend before processing the command. Thonny does this to stop
        the current process and start a new process for the new command.

        This function tries to verify if the backend is restarted by clicking the red button
        or by invoking the `l1test_command`. 
            - if the backend is restarted by clicking the red button, so the treeview is cleaned.
            - if the backend is restarted by invoking the `l1test_command`, so we show in the
            treeview that the l1test is being executed.
            - if the backend is restarted by invoking an other command, so nothing is done.
        
        The problem is that we cannot know who generates the `BackendRestart` event. So to know 
        if the `BackendRestart` event is generated by the `l1test_command` we use the attribute 
        `self._is_l1test_running`. This attribute was setted to True before sending the 
        `l1test_command` to the backend.
        
        Args:
            event (WorkbenchEvent): The event generated from backend restart.
        """
        if (event.get("sequence") == BACKEND_RESTART_EVENT):   
            # Quand le backend est redémarré on efface les exceptions récemment affichée par l1test 
            self.set_has_exception(False)
            self.clean_error_view() 
            
            treeview:L1TestTreeView = self._reporter.get_treeview()
            treeview.disable_menu()
            # L'attribut "full" est un boolean, si c'est "True" alors le backend procède a un
            # redémarrage complet (c'est le cas quand on appuit sur le bouton rouge Stop/Restart).
            # Si c'est False alors c'est un redémarrage partiel du backend (c'est le cas d'un appel 
            # d'une nouvelle commande).
            if event.get("full"):
                self.terminate_running()
                self.clean_treeview(clear_all=True, clear_verdicts_data=True) 
            elif self.is_running(): # si le plugin a été lancé par l'utilisateur
                self.clean_treeview(clear_all=True, clear_verdicts_data=True) 
                treeview.insert_in_header("Starting executing tests ...", clear=True, tags="gray", 
                                            image=PENDING_ICON)
            else: # probablement une autre commande a déclenché le Restart du backend -> on fait rien
                pass
            self.hide_errorview_and_show_treeview()
    
    def _handle_execution_state(self, msg: InlineResponse):
        """
        This function is called when an event of type InlineResponse is received. 
        This function verifies the source of the event. If the source is L1Test so it will access to the 
        received response and then it will handle the state of the execution.
        """
        if self._is_relevant_response(msg):
            state:str = msg.get("state")
            if state == PENDING_STATE:
                lineno = msg.get("lineno") # la ligne en cours d'execution est récupéré seulement si le state == PENDING_STATE
                status_msg = L1TEST_IN_PROGRESS + ": line " + str(lineno) 
                # you can comment the following line if you don't want to show the status message
                self._reporter.get_treeview().insert_in_header(status_msg, clear=True, tags="gray", image=PENDING_ICON)
                self.set_pending(True)
            else: # si le state == FINISHED_STATE
                self.set_pending(False) 
    
    def show_error_view(self, error_msg=None, error_title=None, force=False, show_treeview=False):
        if self._has_exception or force:
            self.show_errors(exception_msg=error_msg, title=error_title)
            self._reporter.get_error_view().show_view()   
            if show_treeview:
                self._reporter.get_treeview().show_view()
            else: 
                self._reporter.get_treeview().hide_view()
    
    def set_and_show_verdicts_on_treeview(self, verdicts:List[L1DocTest]):
        if not self._has_exception:
            self.set_and_show_verdicts(verdicts)     
            self.hide_errorview_and_show_treeview()
    
    def hide_errorview_and_show_treeview(self):
        self._reporter.get_treeview().show_view()
        self.clean_error_view()
        self._reporter.get_error_view().hide_view()
            
    def _handle_clicked_exception(self, event:WorkbenchEvent=None):
        """ 
        This function is called when the user clicks on an exception in the treeview. Then, 
        the error view is displayed with the details of the clicked exception. 
        """
        possible_keys = ["sequence", "error_title", "error_msg", "side_by_side"]
        if event and event.get("sequence") == L1TREE_VIEW_EVENT:
            assert all(key in possible_keys for key in event.__dict__.keys())
            self.set_has_reported_exception(True)
            self.show_error_view(error_msg=event.error_msg, 
                                  error_title=event.error_title, 
                                  force=True,
                                  show_treeview=event.side_by_side)
            
    def set_and_show_verdicts(self, test_results:List[L1DocTest]):
        """
        Report the verdicts on the view.

        Args:
            test_results (dict): The recieved verdicts from the backend.
        """
        self._reporter.display_tests_results(test_results) 
    
    def show_errors(self, exception_msg, title=None):
        """
        Clears the error view and reports the error message on it.
        
        Args:
            exception_msg (str): The error message to display.
            title (str, optional): The title of the error message. Defaults to None.
        """
        title = CANNOT_RUN_TESTS_MSG if title is None else title
        self._reporter.display_error_msg(exception_msg, title=title) 
    
    def _is_relevant_response(self, msg: ToplevelResponse, cmd_name:str=BACKEND_COMMAND):
       """Returns True if the TopLevelResponse is relevant to the l1test plugin."""         
       return msg.get("command_name") == cmd_name
    
    def clean_error_view(self):
        self._reporter.get_error_view().clear()
               
    def clean_treeview(self, clear_verdicts_data=False, clear_all=True, clear_errorview=False): 
        self._reporter.get_treeview().clear_tree(clear_verdicts_data=clear_verdicts_data, 
                                                 clear_all=clear_all, 
                                                 clear_errorview=clear_errorview)
         
    def set_is_running(self, value=True):
        self._is_l1test_running = value
    
    def is_running(self):
        return self._is_l1test_running
    
    def is_pending(self):
        return self._is_pending
    
    def set_pending(self, is_pending:bool):
        self._is_pending = is_pending
    
    def set_has_exception(self, has_exception:bool):
        self._has_exception = has_exception
    
    def has_exception(self):
        return self._has_exception
    
    def terminate_running(self):
        """
        Set the state of `L1TestRunner` as terminated.
        This function sets the `_is_l1test_running` attribute to False
        """
        self.set_pending(False)
        self.set_is_running(False)
        
    def get_reporter(self) -> L1TestReporter:
        return self._reporter
    
    def set_reporter(self, reporter):
        self._reporter = reporter

    def has_reported_exception(self):
        return self._has_reported_exception
    
    def set_has_reported_exception(self, value:bool):
        self._has_reported_exception = value



from .doc_generator import *

# Ceci est une implémentation de création d'un singleton pour L1TestRunner
_doc_generator: DocGenerator = None

def get_doc_generator(auto:bool) -> DocGenerator:
    """
    If there's no `DocGenerator` instance creates one and returns it,
    otherwise returns the current `DocGenerator` instance.
    """
    strategy = AutoGenerationStrategy() if auto else ManualGenerationStrategy()
    if not _doc_generator:
        return DocGenerator(strategy) 
    else:
        _doc_generator.set_strategy(strategy)
        return _doc_generator               


from thonny.plugins.cpython_backend.cp_back import  MainCPythonBackend
from unittest.mock import *

class MockBackend(MainCPythonBackend):
    def __init__(self):
        pass
    
    def send_message(self, msg) -> None: pass

backend_patch = patch("thonnycontrib.backend.l1test_backend.BACKEND", 
                      return_value=MockBackend())


from types import ModuleType
import unittest as ut
from thonnycontrib.backend.ast_parser import L1DocTestFlag
from thonnycontrib.backend.verdicts.PassedSetupVerdict import PassedSetupVerdict
from thonnycontrib.l1test_frontend.l1test_treeview import L1TestTreeView
from thonnycontrib.backend.evaluator import Evaluator
from thonnycontrib.exceptions import *
from thonnycontrib.backend.verdicts.FailedWhenExceptionExpectedVerdict import FailedWhenExceptionExpectedVerdict
from thonnycontrib.backend.verdicts.ExceptionVerdict import ExceptionVerdict
from thonnycontrib.backend.verdicts.FailedVerdict import FailedVerdict
from thonnycontrib.backend.verdicts.PassedVerdict import PassedVerdict
from thonnycontrib.backend.doctest_parser import *
from unittest.mock import *
from thonnycontrib.tests.fixtures.backend_mock import *

# #######################################################
#    Tous les tests qui suivent vérifient le rendu ur la treeview
# #######################################################


# ######## ATTENTION : NOTE IMPORTANTE ##################
# Dans les tests suivants, on assure la présence du verdict pour les setups qui réussissent (PassedSetupVerdict),
# mais en réalité leurs verdicts ne sont pas affichés sur la vue. Ils sont filtrés dans la méthode _add_verdicts_to_node() 
# de la classe L1TestTreeView.
# #######################################################

class MockL1TestTreeView(L1TestTreeView):
    def __init__(self):
        pass
        
class TestTestReporter(ut.TestCase):
    def setUp(self):
        self.l1TestTreeview = MockL1TestTreeView()
        self.evaluator = Evaluator(filename="<string>")
        self.mock_backend = backend_patch.start()
    
    def tearDown(self) -> None:
        del self.evaluator
        backend_patch.stop()
    
    def test_summarize_case_1(self):
        fake_source = \
"""
def f(a, b):
    '''
    $$$ f(1, 1)     # Passed verdict
    2
    $$$ f(1, 1)     # Failed verdict
    0
    $$$ f(20, 20)   # Exception verdict
    0
    $$$ l = []      # It's a setup -> no verdict
    '''
    if a < 0 and b < 0:
        return None
    if a > 10 and b > 10:
        raise ValueError()
    return a + b
"""
        self.evaluator.set_source(fake_source)
        self.evaluator.set_module(self.__build_module_from_source(fake_source))
        
        # #####################################################
        # ------------- Verification des verdicts -------------
        expected_verdicts_colors = {PassedVerdict: "green",
                                    FailedVerdict: "red",
                                    ExceptionVerdict: "red",
                                    PassedSetupVerdict: "black"}
        
        l1doctests = self.evaluator.evaluate()
        
        # on assure qu'il existe un seul l1doctest avec ses examples
        self.assertTrue(len(l1doctests) == 1)
        l1doctest = l1doctests[0]
        self.assertTrue(l1doctest.is_evaluated()) # s'assure que le l1doctest a été évalué

        examples = l1doctest.get_examples() 
        # on s'assure que le nombre de verdicts calculés est exactement le nombre de verdicts attendus
        self.assertEqual(len(examples), len(expected_verdicts_colors)) 
        # on assure que les verdicts récupérés sont exactement les verdicts attendus
        self.assertEqual([e.get_verdict().__class__ for e in examples], list(expected_verdicts_colors.keys()))

        # ##################################################################
        # ------------- Verification des couleurs des verdicts -------------
        self.assertEqual([e.get_verdict().get_color() for e in examples], list(expected_verdicts_colors.values())) 
        
        # ##################################################################
        # ------------------ Verification du SUMMARIZE ---------------------
        summarize = self.l1TestTreeview.build_summarize_object(l1doctests)
        self.assertTrue(summarize.total == 3) 
        self.assertTrue(summarize.success == 1) 
        self.assertTrue(summarize.failures == 1) 
        self.assertTrue(summarize.errors == 1) 
        self.assertTrue(summarize.empty == 0)
        
        # ###########################################################################
        # ------------------ Verification du status du l1doctest ---------------------
        self.assertEqual(l1doctest.get_flag(), L1DocTestFlag.FAILED_FLAG)
        self.assertEqual(l1doctest.get_flag().get_color(), "red")
    
    def test_summarize_case_2(self):
        """
        Dans ce test on vérifie le résultat de summarize quand la docstring 
        contient que des setups.
        """
        fake_source = \
"""
def f(a, b):
    '''
    $$$ a = 0 
    '''
    return a + b
"""
        self.evaluator.set_source(fake_source)
        self.evaluator.set_module(self.__build_module_from_source(fake_source))
        
        # #####################################################
        # ------------- Verification des verdicts -------------        
        l1doctests = self.evaluator.evaluate()
        
        # on s'assure qu'il existe un seul noeud AST 
        self.assertEqual(len(l1doctests), 1)
        
        l1doctest = l1doctests[0]
        self.assertTrue(l1doctest.is_evaluated()) # s'assure que le l1doctest a été évalué
        self.assertEqual([PassedSetupVerdict], [e.get_verdict().__class__ for e in l1doctest.get_examples()]) 
        self.assertEqual(l1doctest.get_flag(), L1DocTestFlag.EMPTY_FLAG)
        self.assertEqual(l1doctest.get_flag().get_color(), "orange")
        
        # ##################################################################
        # ------------------ Verification du SUMMARIZE ---------------------
        summarize = self.l1TestTreeview.build_summarize_object(l1doctests)
        self.assertTrue(summarize.empty == 1)
        self.assertTrue(summarize.total == 0) 
        self.assertTrue(summarize.success == 0) 
        self.assertTrue(summarize.failures == 0) 
        self.assertTrue(summarize.errors == 0) 
            
    def test_summarize_case_3(self):
        """
        Dans ce test on vérifie le résultat de summarize quand la docstring 
        contient à la fois des setups et des tests.
        """
        fake_source = \
"""
def f(a, b):
    '''
    $$$ a = 1 
    $$$ b = 1
    $$$ f(a, b)
    2
    '''
    return a + b
"""
        self.evaluator.set_source(fake_source)
        self.evaluator.set_module(self.__build_module_from_source(fake_source))
        
        # #####################################################
        # ------------- Verification des verdicts -------------
        expected_verdicts = [PassedSetupVerdict, PassedSetupVerdict, PassedVerdict]
        expected_verdicts_colors = ["black", "black", "green"]
        
        l1doctests = self.evaluator.evaluate()
        
        # on assure qu'il existe un seul l1doctest avec ses examples
        self.assertTrue(len(l1doctests) == 1)
        l1doctest = l1doctests[0]
        self.assertTrue(l1doctest.is_evaluated()) # s'assure que le l1doctest a été évalué

        examples = l1doctest.get_examples() 
        # on s'assure que le nombre de verdicts calculés est exactement le nombre de verdicts attendus
        self.assertEqual(len(examples), len(expected_verdicts)) 
        # on assure que les verdicts récupérés sont exactement les verdicts attendus
        self.assertEqual([e.get_verdict().__class__ for e in examples], expected_verdicts)

        # ##################################################################
        # ------------- Verification des couleurs des verdicts -------------
        self.assertEqual([e.get_verdict().get_color() for e in examples], expected_verdicts_colors) 
        
        # ##################################################################
        # ------------------ Verification du SUMMARIZE ---------------------
        summarize = self.l1TestTreeview.build_summarize_object(l1doctests)
        self.assertTrue(summarize.total == 1) 
        self.assertTrue(summarize.success == 1)  
        self.assertTrue(summarize.failures == 0) 
        self.assertTrue(summarize.errors == 0) 
        self.assertTrue(summarize.empty == 0)
        
        # ###########################################################################
        # ------------------ Verification du status du l1doctest ---------------------
        self.assertEqual(l1doctest.get_flag(), L1DocTestFlag.PASSED_FLAG)
        self.assertEqual(l1doctest.get_flag().get_color(), "green")
        
    def test_summarize_case_4(self):
        """
        Dans ce test on vérifie le résultat de summarize quand 
        il existe 2 noeuds ast dont un d'eux contient que des setups 
        et l'autre contient des tests.
        """
        fake_source = \
"""
def minus(a, b):
    '''
    $$$ a = 1
    '''
    return a - b
    
def somme(a, b):
    '''
    $$$ a = 1
    $$$ b = 1
    $$$ somme(a, b)
    2
    '''
    return a + b
"""
        self.evaluator.set_source(fake_source)
        self.evaluator.set_module(self.__build_module_from_source(fake_source))
        
        # #####################################################
        # ------------- Verification des verdicts -------------
        l1doctests = self.evaluator.evaluate()
        
        # on assure qu'il existe 2 l1doctests 
        self.assertTrue(len(l1doctests) == 2)
        
        # ---- vérification pour `minus()` ---
        l1doctest = l1doctests[0]
        self.assertTrue(l1doctest.is_evaluated()) # s'assure que le l1doctest a été évalué
        self.assertTrue("minus" in l1doctest.get_name())
        examples = l1doctest.get_examples()
        self.assertEqual(len(l1doctest.get_examples()), 1) 
        self.assertEqual([PassedSetupVerdict], [e.get_verdict().__class__ for e in examples])
        # le flag est `EMPTY_FLAG` car il n'y a pas de tests pour `minus()`
        self.assertEqual(l1doctest.get_flag(), L1DocTestFlag.EMPTY_FLAG)

        # ---- vérification pour `somme()` ---
        l1doctest = l1doctests[1]
        self.assertTrue(l1doctest.is_evaluated()) # s'assure que le l1doctest a été évalué
        self.assertTrue("somme" in l1doctest.get_name())
        examples = l1doctest.get_examples()
        self.assertEqual(len(examples), 3) 
        self.assertEqual([PassedSetupVerdict, PassedSetupVerdict, PassedVerdict], 
                         [e.get_verdict().__class__ for e in examples])
        self.assertEqual(["black", "black", "green"], 
                         [e.get_verdict().get_color() for e in examples])
        # le flag est `PASSED_FLAG` car il existe un test qui passe pour `somme()`
        self.assertEqual(l1doctest.get_flag(), L1DocTestFlag.PASSED_FLAG)
        
        # ##################################################################
        # ------------------ Verification du SUMMARIZE ---------------------
        summarize = self.l1TestTreeview.build_summarize_object(l1doctests)
        self.assertTrue(summarize.total == 1) 
        self.assertTrue(summarize.success == 1) 
        self.assertTrue(summarize.failures == 0) 
        self.assertTrue(summarize.errors == 0) 
        self.assertTrue(summarize.empty == 1) 
    
    def test_summarize_case_5(self):
        """
        Dans ce test on vérifie le résultat de summarize quand 
        l'invite est `$$e`.
        """
        fake_source = \
"""
class NotAnException: pass

def somme(a, b):
    '''
    $$e somme(10, 10)       # le cas qui échoue(failed verdict)
    ValueError
    
    $$$ a, b = -1, -1
    $$e somme(a, b)         # le cas qui passe
    ValueError
    $$e somme(a, b)         
    UndefinedException      # le want déclare une exception qui n'existe pas
    $$e somme(a, b)
    Invalid name            # le want contient une erreur de nommage
    $$e somme(a, b)
    NotAnException          # le want n'est pas une exception qui hérite de BaseException
    '''
    if a > 9  or b > 9:
        raise Exception("a et b doivent être que des chiffres")
    if a < 0 and b < 0:
        raise ValueError()
    return a + b
"""
        self.evaluator.set_source(fake_source)
        self.evaluator.set_module(self.__build_module_from_source(fake_source))
        
        # #####################################################
        # ------------- Verification des verdicts -------------
        l1doctests = self.evaluator.evaluate()
        
        # on s'assure qu'il existe deux noeuds AST : la fonction `somme()` et la classe `NotAnException`
        self.assertEqual(len(l1doctests), 2)
        
        # ---- vérification pour le l1Doctest correpondant à la classe `NotAnException` ---
        l1doctest = l1doctests[0] # le l1doctest de la classe `NotAnException`
        self.assertTrue(l1doctest.is_evaluated()) # s'assure que le l1doctest a été évalué
        self.assertEqual(l1doctest.get_flag(), L1DocTestFlag.EMPTY_FLAG)
        self.assertEqual(l1doctest.get_flag().get_color(), "orange")
        
        # ---- vérification pour le l1Doctest correpondant à la fonction `somme()` ---
        expected_verdicts = [FailedWhenExceptionExpectedVerdict, PassedSetupVerdict, PassedVerdict] + \
                            [ExceptionVerdict]*3  # there are 3 exceptions verdicts
        expected_colors = ["red", "black", "green"] + ["red"]*3
        
        l1doctest = l1doctests[1] # le l1doctest de la fonction `somme()`
        self.assertTrue(l1doctest.is_evaluated()) # s'assure que le l1doctest a été évalué
        self.assertEqual(l1doctest.get_flag(), L1DocTestFlag.FAILED_FLAG)
        self.assertEqual(l1doctest.get_flag().get_color(), "red")
        
        examples = l1doctest.get_examples()
        self.assertEqual(expected_verdicts, [e.get_verdict().__class__ for e in examples])
        self.assertEqual(expected_colors, [e.get_verdict().get_color() for e in examples])
        
        # ##################################################################
        # ------------------ Verification du SUMMARIZE ---------------------
        summarize = self.l1TestTreeview.build_summarize_object(l1doctests)  
        self.assertTrue(summarize.total == 5) 
        self.assertTrue(summarize.success == 1)  
        self.assertTrue(summarize.failures == 1) 
        self.assertTrue(summarize.errors == 3) 
        self.assertTrue(summarize.empty == 1)

             
    def __build_module_from_source(self, source: str) -> ModuleType:
        """
        Build a module containing the functions declared in the given `source`.
        """
        from types import ModuleType
        fake_module = ModuleType(self.evaluator.get_filename())
        exec(source, fake_module.__dict__)
        return fake_module

if __name__ == '__main__':
    ut.main(verbosity=2)   
        



# ########################################## #
#              Backend exceptions            #
# ########################################## #
class BackendException(BaseException):
    """
        This is the superclass of all the exceptions raised in the
        l1test backend. 
    """
    def get_prefix_message(self) -> str:
        """
            Returns the title to be shown in the error view. 
            The current title is "Cannot run L1Test : ". 

            If it returns `None` so the old title will be kept. 
            Otherwise, it will be changed to the returned value.
        """
        return None

class EmptyValueError(BackendException, ValueError):
    """
    Exception raised when a value is empty. 
    """
    pass

class CompilationError(BackendException): 
    """
    Exception raised when the parsing by the AST parser, `exec()` or by the `compile()` is failed.
    """
    pass

class NoTestFoundException(BackendException):
    """
    Exception raised when an there's no test in the source code. 
    """
    def get_prefix_message(self):
        return ""   
    
class NoFunctionSelectedToTestException(BackendException):
    """
    Exception raised when the selected line doesn't contain a function/class signature.
    """
    pass 

class CannotImportModuleException(BackendException):
    """
    Exception raised when a module cannot be imported by the importlib library.
    """
    pass

class RuntimeException(BackendException):
    """
    Exception raised when a runtime error is raised by the `exec()` method
    """
    pass

class DoctestParserException(BackendException):
    """
    This should be a superclass of all the exceptions raised by the doctest parser
    """
    def __init__(self, filename, lineno, message):
        super().__init__('File %s, line %s\n%s' % (filename, lineno, message))

class SpaceMissingAfterPromptException(DoctestParserException):
    """
    Exception raised when a <space> charatcer is missing after the L1test symbol ($$$) 
    """
    def __init__(self, filename, lineno, invite_sans_blanc):
        super().__init__(filename, lineno, 
                    'A space is missing\n after the prompt : %s' % (invite_sans_blanc)) 
            
            
# ########################################## #
#              Frontend exceptions           #
# ########################################## #
class FrontendException(BaseException):
    """
        This is the superclass of all the exceptions not raised in the
        l1test backend. 
    """
    pass

class NoEditorFoundException(FrontendException):
    """
    exception which includes all exceptions related to the code editor.
    """
    pass
        
class NotSavedFileException(FrontendException):
    """
    Exception raised when the source code is not saved yet on the disk. 
    """
    pass

class EmptyEditorException(FrontendException):
    """
    Exception raised when an editor is still empty. 
    """
    pass

class NoFunctionSelectedToDocumentException(FrontendException):
    """
    Exception raised when no function is selected to document. 
    """
    pass

class DocGeneratorParserException(FrontendException):
    """
    Exception raised when no docstring cannot be generated by the generator.
    This exception is raised when a function signature contains a compilation
    error.
    """
    pass


# Module doctest.
# Released to the public domain 16-Jan-2001, by Tim Peters (tim@python.org).
# Major enhancements and refactoring by:
#     Jim Fulton
#     Edward Loper

# Provided as-is; use at your own risk; no warranty; no promises; enjoy!

# modified pour les besoins de L1test
# - modification de DoctestParser : expression régulière _EXAMPLE_RE changée
# pour supporter plusieurs invites différentes, et forcer à ce que les lignes
# source et want soient sur une seule ligne (ou absente pour want)
# - _check_prompt_blank : cette fonction lèvait une ValueError qd il manque
# un caractère espace après l'invite, ce qui déclenchait l'ouverture d'une
# pop-up "internal error" assez affreuse. Maintenant l'erreur levée est
# ManqueUnEspace et le message associé est affiché ds la fenêtre de test.
__docformat__ = 'reStructuredText en'

__all__ = [
    # 0, Option Flags
    'register_optionflag',
    'DONT_ACCEPT_TRUE_FOR_1',
    'DONT_ACCEPT_BLANKLINE',
    'NORMALIZE_WHITESPACE',
    'ELLIPSIS',
    'SKIP',
    'IGNORE_EXCEPTION_DETAIL',
    'COMPARISON_FLAGS',
    'REPORT_UDIFF',
    'REPORT_CDIFF',
    'REPORT_NDIFF',
    'REPORT_ONLY_FIRST_FAILURE',
    'REPORTING_FLAGS',
    'FAIL_FAST',
    # 1. Utility Functions
    # 2. Example & DocTest
    'Example',
    # 3. Doctest Parser
    'DocTestParser',
]
from abc import *
import __future__
import re
from collections import namedtuple
import sys, ast

# ajout L1test
from thonnycontrib.properties import L1TEST_EXCEPTION_SYMBOL, L1TEST_SYMBOL1, L1TEST_SYMBOL2, L1TEST_SYMBOL3

from .verdicts.FailedVerdict import *
from .verdicts.FailedWhenExceptionExpectedVerdict import *
from .verdicts.PassedVerdict import *
from .verdicts.ExceptionVerdict import *
from .verdicts.PassedSetupVerdict import *

from thonnycontrib.exceptions import *
from thonnycontrib.utils import *

TestResults = namedtuple('TestResults', 'failed attempted')

# There are 2 basic classes:
#  - Example: a <source, want> pair, plus an intra-docstring line number.
#  - DocTest: a collection of examples, parsed from a docstring, plus
#    info about where the docstring came from (name, filename, lineno).
#  - DocTestFinder: extracts DocTests from a given object's docstring and
#    its contained objects' docstrings.
#  - DocTestRunner: runs DocTest cases, and accumulates statistics.
#
# So the basic picture is:
#
#                             list of:
# +------+                   +---------+                   +-------+
# |object| --DocTestFinder-> | DocTest | --DocTestRunner-> |results|
# +------+                   +---------+                   +-------+
#                            | Example |
#                            |   ...   |
#                            | Example |
#                            +---------+

# Option constants.

OPTIONFLAGS_BY_NAME = {}
def register_optionflag(name):
    # Create a new flag unless `name` is already known.
    return OPTIONFLAGS_BY_NAME.setdefault(name, 1 << len(OPTIONFLAGS_BY_NAME))

DONT_ACCEPT_TRUE_FOR_1 = register_optionflag('DONT_ACCEPT_TRUE_FOR_1')
DONT_ACCEPT_BLANKLINE = register_optionflag('DONT_ACCEPT_BLANKLINE')
NORMALIZE_WHITESPACE = register_optionflag('NORMALIZE_WHITESPACE')
ELLIPSIS = register_optionflag('ELLIPSIS')
SKIP = register_optionflag('SKIP')
IGNORE_EXCEPTION_DETAIL = register_optionflag('IGNORE_EXCEPTION_DETAIL')

COMPARISON_FLAGS = (DONT_ACCEPT_TRUE_FOR_1 |
                    DONT_ACCEPT_BLANKLINE |
                    NORMALIZE_WHITESPACE |
                    ELLIPSIS |
                    SKIP |
                    IGNORE_EXCEPTION_DETAIL)

REPORT_UDIFF = register_optionflag('REPORT_UDIFF')
REPORT_CDIFF = register_optionflag('REPORT_CDIFF')
REPORT_NDIFF = register_optionflag('REPORT_NDIFF')
REPORT_ONLY_FIRST_FAILURE = register_optionflag('REPORT_ONLY_FIRST_FAILURE')
FAIL_FAST = register_optionflag('FAIL_FAST')

REPORTING_FLAGS = (REPORT_UDIFF |
                   REPORT_CDIFF |
                   REPORT_NDIFF |
                   REPORT_ONLY_FIRST_FAILURE |
                   FAIL_FAST)

# Special string markers for use in `want` strings:
BLANKLINE_MARKER = '<BLANKLINE>'
ELLIPSIS_MARKER = '...'


######################################################################
## 2. Example 
######################################################################
## - An "example" is a <source, want> pair, where "source" is a
##   fragment of source code, and "want" is the expected output for
##   "source."  The Example class also includes information about
##   where the example was extracted from.

class Example(ABC):
    """
    A single doctest example, consisting of source code and expected
    output.  `Example` defines the following attributes:

      - source: A single Python statement, always ending with a newline.
        The constructor adds a newline if needed.

      - want: The expected output from running the source code (either
        from stdout, or a traceback in case of exception).  `want` ends
        with a newline unless it's empty, in which case it's an empty
        string.  The constructor adds a newline if needed.

      - exc_msg: The exception message generated by the example, if
        the example is expected to generate an exception; or `None` if
        it is not expected to generate an exception.  This exception
        message is compared against the return value of
        `traceback.format_exception_only()`.  `exc_msg` ends with a
        newline unless it's `None`.  The constructor adds a newline
        if needed.

      - lineno: The line number within the DocTest string containing
        this Example where the Example begins.  This line number is
        zero-based, with respect to the beginning of the DocTest.

      - indent: The example's indentation in the DocTest string.
        I.e., the number of space characters that precede the
        example's first prompt.

      - options: A dictionary mapping from option flags to True or
        False, which is used to override default options for this
        example.  Any option flags not contained in this dictionary
        are left at their default value (as specified by the
        DocTestRunner's optionflags).  By default, no options are set.
    """
    def __init__(self, filename:str, source:str, want:str, invite:str, exc_msg:str=None, lineno:int=0, indent:int=0, 
                 options:dict=None):
        # Normalize inputs.
        if not source.endswith('\n'):
            source += '\n'
        if want and not want.endswith('\n'):
            want += '\n'
        if exc_msg is not None and not exc_msg.endswith('\n'):
            exc_msg += '\n'
        # Store properties.
        self.filename = filename
        self.source = source
        self.want = want
        self.lineno = lineno
        self.indent = indent
        self.invite = invite
        if options is None: options = {}
        self.options = options
        self.exc_msg = exc_msg
        
        # the verdict of this example. By default is an empty verdict.
        self.__verdict:ExampleVerdict = None

    @abstractmethod
    def is_a_test(self) -> bool:
        """
        Returns:
            bool: Return True if the example is considered to have an oracle. 
            Return False if the example is considered as a setup.
        """
        pass
    
    @abstractmethod
    def _exec_and_computes_verdict(self, globals:dict={}):
        '''Executes the example and returns a verdict, eg FailedTest,
        PassedTest...

        Args : 
        - globals : the dict of globals variables
        '''
        pass
    
    def compute_and_set_verdict(self, globals:dict={}):
        '''
        This is the main method to evaluate and compute the verdict 
        of an example. This method invokes the abstract method `_exec_and_computes_verdict`, 
        then set the returned verdict to this example.
        
        Args : 
        - globals : the dict of globals variables
        '''
        self.__verdict = self._exec_and_computes_verdict(globals)
        return self.__verdict
    
    def _execute_statement(self, source:str, globals:dict, lineno:int):
        """
        Execute a source code using the given globals. 
        
        This function should be invoked after a `compile()`. The compile() function is removed from 
        this function because it's hard to know if an exception is raised by a compilation error
        or by an executed statement(using exec()). 
        
        Args:
            source (str): A string representing a source code. 
            globals (dict): The dictionary of the global variables.
            lineno (int): The line of the test to execute.
        Raises:
            RuntimeException: When the source code raises a runtime error(ex. NameError...)
        """
        try:
            exec(source, globals)
        except BaseException as e:
            error_info = sys.exc_info()
            formatted_error = replace_filename(self.filename, get_last_exception(error_info))
            
            # For runtime errors the line mentionned in the error raised by `exec()` is always equal to 1.
            # To learn about runtime errors in python: https://www.geeksforgeeks.org/runtime-errors/
            has_error_lineno = re.search(r'line (\d+)', formatted_error)
            if has_error_lineno:
                error_line = int(has_error_lineno.group(1))
                # comme cette fonction est toujours appelée pour executer un test ou sa valeur attendu
                # alors on aura jamais la ligne d'erreur est égal à 1. Donc, on remplace la ligne d'erreur
                # par le numéro de ligne du test.
                if error_line == 1: 
                    formatted_error = replace_error_line(formatted_error, lineno)
            raise RuntimeException(formatted_error)
    
    def _check_syntax(self, source:str, lineno:int, mode="single"):
        """
        Compile the `source` of a test statement and raise an exception when a 
        compilation error occurs.

        Args:
            source (str): The statement to be compiled
            mode (str): Defaults to "single".
        Raises:
            CompilationError: when a compilation error occurs.
        """
        try:
            compile(source, filename=self.filename, mode=mode)
        except Exception:
            formatted_error = get_last_exception(sys.exc_info())
            raise CompilationError(replace_error_line(formatted_error, lineno))
    
    def _create_affectation_statement(self, statement1, statement2):
        """
        Creates an affectation expression by affecting the statement2 to statement1.

        The result will be : "%s = %s" % (statement1, statement2)
        
        Args:
            statement1 (str): the left side of the affectation expression
            statement2 (str): the right side of the affectation expression

        Returns:
            (str): the affectation expression using `statement1` and `statement1`. 
        """
        return "%s = (%s)" % (statement1, statement2)
    
    def get_verdict(self):
        return self.__verdict
    
    def is_evaluated(self):
        """Returns True if the example has a verdict, False otherwise."""
        return self.__verdict is not None
        
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented

        return self.filename == other.filename and \
               self.source == other.source and \
               self.want == other.want and \
               self.lineno == other.lineno and \
               self.indent == other.indent and \
               self.options == other.options and \
               self.exc_msg == other.exc_msg and \
               self.__verdict == other.get_verdict()

    def __hash__(self):
        return hash((self.filename, self.source, self.want, self.lineno, 
                     self.indent, self.exc_msg))

    
class ExampleWithoutExpected(Example):
    """
    An `ExampleWithNoExpected` is an example without a `want` and cannot be considered as a test.
    """
    def _exec_and_computes_verdict(self, globals:dict={}):
        try:  
            # we chech the syntax of the test before its execution
            self._check_syntax(self.source, self.lineno) 
            self._execute_statement(self.source, globals, self.lineno)
            return PassedSetupVerdict(self.filename, tested_line=self.source, lineno=self.lineno)
        except (RuntimeException, CompilationError) as error:
            return ExceptionVerdict(self.filename, tested_line=self.source.strip(), 
                                 expected_result=self.want.strip("\n"), lineno=self.lineno, 
                                 error_details=str(error))
            
    def is_a_test(self) -> bool:
        return False

class ExampleExceptionExpected(Example):
    """
    An `ExampleExceptionExpected` is a test example that expects an exception to be raised.
    """    
    def _exec_and_computes_verdict(self, globals:dict={}):
        want = self.want.strip("\n").strip()
        try:
            self.__check_is_an_exception(self.want.strip("\n"))
            # On vérifie si l'exception hérite de la BaseException
            self.__is_exception_inherits_from_BaseException(globals, want)
            return self.__execute_and_get_verdict(globals, want)  
        except (ValueError, TypeError, NameError) as e:
            lineno = self.lineno if isinstance(e, EmptyValueError) else self.lineno + 1
            # get the super class of EmptyValueError 
            error_class_name = ValueError.__name__ if isinstance(e, EmptyValueError) else e.__class__.__name__
            error_msg = "%s:\n  %s" % (error_class_name, str(e))
            return self.__build_exception_verdict(error_msg, want, lineno)         
    
    def __execute_and_get_verdict(self, globals, want:str):
        try:
            # il faudrait quand même compiler la source avant de l'executer
            self._check_syntax(self.source, self.lineno)
            
            exec(self.source, globals)
            
            # si on est ici alors aucune exc levée -> FailedTest
            # <nom_exc> was not raised by <source>
            return FailedWhenExceptionExpectedVerdict(self.filename, tested_line=self.source.strip(), 
                                expected_result=want, lineno=self.lineno)
            
        except BaseException as e: # ici on a forcément une exception levée par le code source
                # mais a priori rien de syntaxique car on est sur un seul nom
            error_type, _, _ = sys.exc_info()    
            exc_name_effective = str(error_type.__name__)
            if want == exc_name_effective: 
                return PassedVerdict(self.filename, tested_line=self.source.strip(), 
                                    expected_result=want, lineno=self.lineno)   
            else: # forcément une erreur de compilation ou encore une erreur de runtime
                # si c'est une erreur `CompilationError` alors on récupère seulement sa représentation      
                if isinstance(e, CompilationError):   # if CompilationError -> ExceptionTest verdict
                    return ExceptionVerdict(self.filename, tested_line=self.source.strip(), 
                                            expected_result=want, lineno=self.lineno, error_details=str(e))
                else: # sinon on doit chercher la dernière exception. 
                    failure_msg = "Instead, it raises : %s()" % (e.__class__.__name__) 
                    return FailedWhenExceptionExpectedVerdict(self.filename, tested_line=self.source.strip(), 
                                expected_result=want,lineno=self.lineno, failure_message=failure_msg)     

    def __build_exception_verdict(self, exception_msg, want, lineno=None):
        """
        Returns an exception verdict with the given exception message. 

        Args:
            exception_msg (str): The exception message
            want (str): The want of the Example.

        Returns:
            ExceptionTest: an exception verdict with the given exception message. 
        """
        if lineno is None:
            lineno = self.lineno
        common_msg = 'File "%s", line %s\n'%(self.filename, lineno) 
        error_msg = common_msg + exception_msg
        return ExceptionVerdict(self.filename, tested_line=self.source.strip(), 
                                expected_result=want, lineno=lineno, 
                                error_details=error_msg)  
        
    def __check_is_an_exception(self, want:str):
        """Check if the want is a valid exception name that contains letters, 
        numbers and/or underscore.

        Args:
            want (str): a string.

        Raises:
            ValueError: when the want doesn't respect the syntaxe.
        """
        match = re.match("^[a-zA-Z][a-zA-Z0-9_]*", want)
        if not match:
            if not want:
                raise EmptyValueError("The expected exception cannot be empty or blank.")
            else: 
                raise ValueError('The expected exception `%s` is not a valid exception name.' % want.strip())
    
    def __get_exception_from_name(self, globals:dict, exception_name:str):
        """Returns the exception class from its name.

        Args:
            globals (dict): the globals contains all the declared meta informations on the module.
            exception_name (str): the name of the exception.

        Returns:
            (object|None): an object(maybe an exception or not) or None if the exception is not
                        a known exception.
        """
        builtins = globals['__builtins__'].__dict__ if isinstance(globals['__builtins__'], ModuleType) \
                                                    else globals['__builtins__']
        if exception_name in builtins:
            return builtins[exception_name]
        else:  
            # si on est là alors le nom de l'exception n'est pas trouvée dans le __builtins__
            # on doit chercher si l'exception est déclarée dans le fichier executé
            return globals[exception_name] if exception_name in globals else None 
                
    def __is_exception_inherits_from_BaseException(self, globals: dict, exception_name: str) -> None:
        """Checks the superclass of the given exception name. 
        This function first gets the corresponding exception class from its name.
        Then, it verifies if the exception class is really an exception (inherits from `BaseException`).
        
        Args:
            globals (dict): the globals contains all the declared meta informations on the module.
            exception_name (str): the name of the exception.

        Returns:
            None: Return None if the given exception name is an exception that inherits form `BaseException`.
                Otherwise, raises exceptions.
        Raises:
            - TypeError: if the retrieved exception from the given exception name is not an exception that
                        inherits from `BaseException`. 
            - NameError: if cannot retrieve the exception from the given exception name.
        """
        exception: object = self.__get_exception_from_name(globals, exception_name)
        import inspect
        if exception:
            try:
                super_cls = inspect.getmro(exception)[1:] # le premier élément est la classe elle même
                if BaseException in super_cls:
                    return None
                else:
                    raise TypeError("The expected exception `%s` is not a class of type exception." % exception_name) 
            except:
                raise TypeError("The expected exception `%s` is not a class of type exception." % exception_name) 
        raise NameError("The expected exception `%s` cannot be found." % exception_name)

    def is_a_test(self) -> bool:
        return True
    
class ExampleWithExpected(Example):
    """
    An `ExampleWithExpected` is a test example that expects a value to be returned.
    """
    def is_a_test(self) -> bool:
        return True
    
    def _exec_and_computes_verdict(self, globals:dict={}):
        executed, wanted = add_random_suffix("executed"), add_random_suffix("wanted")
        want = self.want.strip("\n")   
        try:             
            # we should check the syntax of the `source` before executing it   
            self._check_syntax(self.source, self.lineno)
            # we should check the syntax of the `want` before executing it 
            self._check_syntax(self.want, self.lineno+1)
            
            # si on est là alors la syntaxe de la source et du want est valide
            # et on doit pouvoir les executer sans problème
            self._execute_statement(self._create_affectation_statement(executed, self.source), 
                                    globals, self.lineno)
            self._execute_statement(self._create_affectation_statement(wanted, self.want), 
                                    globals, self.lineno+1)
            
            self._check_syntax(self.str_comparaison(self.source, self.want), self.filename, "single")
            
            if globals[executed] == globals[wanted]:
                verdict = PassedVerdict(self.filename, tested_line=self.source.strip(), 
                                      expected_result=want, lineno=self.lineno)
            else:
                verdict = FailedVerdict(self.filename, tested_line=self.source.strip(), 
                                      expected_result=want, real_result=globals[executed], 
                                      lineno=self.lineno)
        except (CompilationError, RuntimeException) as e:
            verdict = ExceptionVerdict(self.filename, tested_line=self.source.strip(), 
                                    expected_result=want, lineno=self.lineno, error_details=str(e))
        return verdict

    def str_comparaison(self, expected, got):
        return  "(%s) == (%s)" % (expected, got) 
    
    
######################################################################
## 3. DocTestParser
######################################################################

class DocTestParser:
    """
    A class used to parse strings containing doctest examples.
    """
    # This regular expression is used to find doctest examples in a
    # string.  It defines three groups: `source` is the source code
    # (including leading indentation and prompts); `indent` is the
    # indentation of the first (PS1) line of the source code; and
    # `want` is the expected output (including leading indentation).
    _EXAMPLE_RE = re.compile(r'''
        # Source consists of a PS1 line 
        (?P<source>
            (?:^(?P<indent> [ ]*) (?P<invite> %s|%s|%s|%s) .*))    # PS1 line
        \n?
        # Want consists of zero or one non-blank lines that do not start with PS1.
        (?P<want> (?:(?![ ]*$)    # Not a blank line
                     (?![ ]*(%s|%s|%s|%s))  # Not a line starting with PS1
                     .+$\n?       # But any other line
                  )?) # 0 ou 1 é
        '''%(L1TEST_SYMBOL1, L1TEST_SYMBOL2, L1TEST_SYMBOL3, L1TEST_EXCEPTION_SYMBOL,   # source
             L1TEST_SYMBOL1, L1TEST_SYMBOL2, L1TEST_SYMBOL3, L1TEST_EXCEPTION_SYMBOL),  # want
             re.MULTILINE | re.VERBOSE)

    # version d'origine doctest avec nouvelle invite
    # _EXAMPLE_RE = re.compile(r'''
    #     # Source consists of a PS1 line followed by zero or more PS2 lines.
    #     (?P<source>
    #         (?:^(?P<indent> [ ]*) (%s|%s|%s)    .*)    # PS1 line
    #         (?:\n           [ ]*  \.\.\. .*)*)  # PS2 lines
    #     \n?
    #     # Want consists of any non-blank lines that do not start with PS1.
    #     (?P<want> (?:(?![ ]*$)    # Not a blank line
    #                  (?![ ]*(%s|%s|%s))  # Not a line starting with PS1
    #                  .+$\n?       # But any other line
    #               )*)
    #     '''%(L1TEST_SYMBOL1, L1TEST_SYMBOL2, L1TEST_SYMBOL3, L1TEST_SYMBOL1, L1TEST_SYMBOL2, L1TEST_SYMBOL3), re.MULTILINE | re.VERBOSE)

    
    # A regular expression for handling `want` strings that contain
    # expected exceptions.  It divides `want` into three pieces:
    #    - the traceback header line (`hdr`)
    #    - the traceback stack (`stack`)
    #    - the exception message (`msg`), as generated by
    #      traceback.format_exception_only()
    # `msg` may have multiple lines.  We assume/require that the
    # exception message is the first non-indented line starting with a word
    # character following the traceback header line.
    _EXCEPTION_RE = re.compile(r"""
        # Grab the traceback header.  Different versions of Python have
        # said different things on the first traceback line.
        ^(?P<hdr> Traceback\ \(
            (?: most\ recent\ call\ last
            |   innermost\ last
            ) \) :
        )
        \s* $                # toss trailing whitespace on the header.
        (?P<stack> .*?)      # don't blink: absorb stuff until...
        ^ (?P<msg> \w+ .*)   #     a line *starts* with alphanum.
        """, re.VERBOSE | re.MULTILINE | re.DOTALL)

    # A callable returning a true value iff its argument is a blank line
    # or contains a single comment.
    _IS_BLANK_OR_COMMENT = re.compile(r'^[ ]*(#.*)?$').match

    def parse(self, string, name='<string>'):
        """
        Divide the given string into examples and intervening text,
        and return them as a list of alternating Examples and strings.
        Line numbers for the Examples are 0-based.  The optional
        argument `name` is a name identifying this string, and is only
        used for error messages.
        
        Raises:
            SpaceMissingAfterPromptException: when a space is missing after the prompt.
        """
        string = string.expandtabs()
        # If all lines begin with the same indentation, then strip it.
        min_indent = self._min_indent(string)
        if min_indent > 0:
            string = '\n'.join([l[min_indent:] for l in string.split('\n')])

        output = []
        charno, lineno = 0, 0
        # Find all doctest examples in the string:
        for m in self._EXAMPLE_RE.finditer(string):            
            invite = m.group("invite")

            # Add the pre-example text to `output`.
            output.append(string[charno:m.start()])
            # Update lineno (lines before this example)
            lineno += string.count('\n', charno, m.start())
            # Extract info from the regexp match.
            (source, options, want, exc_msg) = \
                self._parse_example(m, name, invite, lineno)
            
            # Create an Example, and add it to the list.
            if not self._IS_BLANK_OR_COMMENT(source):
                source = re.sub(r'#.*$', "", source).rstrip() # ne pas considérer les commentaires
                example = ExampleFactory(invite, name, source, want, exc_msg,
                                         lineno=lineno+1, indent=min_indent+len(m.group('indent')),
                                         options=options)
                       
                output.append(example)
            # Update lineno (lines inside this example)
            lineno += string.count('\n', m.start(), m.end())
            # Update charno.
            charno = m.end()
        # Add any remaining post-example text to `output`.
        output.append(string[charno:])
        return output

    def get_examples(self, string, name='<string>'):
        """
        Extract all doctest examples from the given string, and return
        them as a list of `Example` objects.  Line numbers are
        0-based, because it's most common in doctests that nothing
        interesting appears on the same line as opening triple-quote,
        and so the first interesting line is called \"line 1\" then.

        The optional argument `name` is a name identifying this
        string, and is only used for error messages.
        """
        return [x for x in self.parse(string, name)
                if isinstance(x, Example)]

    def _parse_example(self, m, name, invite, lineno):
        """
        Given a regular expression match from `_EXAMPLE_RE` (`m`),
        return a pair `(source, want)`, where `source` is the matched
        example's source code (with prompts and indentation stripped);
        and `want` is the example's expected output (with indentation
        stripped).

        `name` is the string's name, and `lineno` is the line number
        where the example starts; both are used for error messages.
        """
        # Get the example's indentation level.
        indent = len(m.group('indent'))

        # Divide source into lines; check that they're properly
        # indented; and then strip their indentation & prompts.
        source_lines = m.group('source').split('\n')
        self._check_prompt_blank(source_lines, indent, invite, name, lineno)
        self._check_prefix(source_lines[1:], ' '*indent + '.', name, lineno)
        source = '\n'.join([sl[indent+4:] for sl in source_lines])

        # Divide want into lines; check that it's properly indented; and
        # then strip the indentation.  Spaces before the last newline should
        # be preserved, so plain rstrip() isn't good enough.
        want = m.group('want')
        want_lines = want.split('\n')
        if len(want_lines) > 1 and re.match(r' *$', want_lines[-1]):
            del want_lines[-1]  # forget final newline & spaces after it
        self._check_prefix(want_lines, ' '*indent, name,
                           lineno + len(source_lines))
        want = '\n'.join([wl[indent:] for wl in want_lines])

        # If `want` contains a traceback message, then extract it.
        m = self._EXCEPTION_RE.match(want)
        if m:
            exc_msg = m.group('msg')
        else:
            exc_msg = None

        # Extract options from the source.
        options = self._find_options(source, name, lineno)

        return source, options, want, exc_msg

    # This regular expression looks for option directives in the
    # source code of an example.  Option directives are comments
    # starting with "doctest:".  Warning: this may give false
    # positives for string-literals that contain the string
    # "#doctest:".  Eliminating these false positives would require
    # actually parsing the string; but we limit them by ignoring any
    # line containing "#doctest:" that is *followed* by a quote mark.
    _OPTION_DIRECTIVE_RE = re.compile(r'#\s*doctest:\s*([^\n\'"]*)$',
                                      re.MULTILINE)

    def _find_options(self, source, name, lineno):
        """
        Return a dictionary containing option overrides extracted from
        option directives in the given source string.

        `name` is the string's name, and `lineno` is the line number
        where the example starts; both are used for error messages.
        """
        options = {}
        # (note: with the current regexp, this will match at most once:)
        for m in self._OPTION_DIRECTIVE_RE.finditer(source):
            option_strings = m.group(1).replace(',', ' ').split()
            for option in option_strings:
                if (option[0] not in '+-' or
                        option[1:] not in OPTIONFLAGS_BY_NAME):
                    raise ValueError('line %r of the doctest for %s '
                                     'has an invalid option: %r' %
                                     (lineno+1, name, option))
                flag = OPTIONFLAGS_BY_NAME[option[1:]]
                options[flag] = (option[0] == '+')
        if options and self._IS_BLANK_OR_COMMENT(source):
            raise ValueError('line %r of the doctest for %s has an option '
                             'directive on a line with no example: %r' %
                             (lineno, name, source))
        return options

    # This regular expression finds the indentation of every non-blank
    # line in a string.
    _INDENT_RE = re.compile(r'^([ ]*)(?=\S)', re.MULTILINE)

    def _min_indent(self, s):
        "Return the minimum indentation of any non-blank line in `s`"
        indents = [len(indent) for indent in self._INDENT_RE.findall(s)]
        if len(indents) > 0:
            return min(indents)
        else:
            return 0

    def _check_prompt_blank(self, lines, indent, invite, name, lineno):
        """
        Given the lines of a source string (including prompts and
        leading indentation), check to make sure that every prompt is
        followed by a space character.  If any line is not followed by
        a space character, then raise ValueError.
        """
        for i, line in enumerate(lines):
            invite_length = len(invite)
            if len(line) >= indent+invite_length+1 and line[indent+invite_length] != ' ':
                # old doctest version
                # raise ValueError('line %r of the docstring for %s '
                #                  'lacks blank after %s: %r' %
                #                  (lineno+i+1, name,
                #                    line[indent:indent+3], line))
                raise SpaceMissingAfterPromptException(name, lineno+i+1, line[indent:])
            
    def _check_prefix(self, lines, prefix, name, lineno):
        """
        Check that every line in the given list starts with the given
        prefix; if any line does not, then raise a ValueError.
        """
        for i, line in enumerate(lines):
            if line and not line.startswith(prefix):
                raise ValueError('line %r of the docstring for %s has '
                                 'inconsistent leading whitespace: %r' %
                                 (lineno+i+1, name, line))
     
def ExampleFactory(invite:str, filename:str, source:str, want:str, exc_msg:str, lineno:int, indent, options) -> Example:
    """
    Creates the appropriate `Example` type depending on the invite command and
    depending on the want value.
    
    - If the invite command is an exception invite so an `ExceptionExample` will be returned.
    - If the want value is empty so a `ExampleWithNoExpected` will be returned.
    - Otherwise, an `ExpectedExample` is returned.
    
    To know more about each `Example` type, please see thier documentations.
    
    Returns:
        Example: an `Example` type depending on the invite command and the want value.
    """
    exception_invite = re.search(L1TEST_EXCEPTION_SYMBOL, invite)

    if exception_invite: 
        example = ExampleExceptionExpected(filename, source, want, invite, exc_msg,
                                            lineno=lineno,
                                            indent=indent,
                                            options=options)
    else:
        if not want:
            example = ExampleWithoutExpected(filename, source, want, invite, exc_msg,
                                            lineno=lineno,
                                            indent=indent,
                                            options=options)
        else:
            example = ExampleWithExpected(filename, source, want, invite, exc_msg,
                                            lineno=lineno,
                                            indent=indent,
                                            options=options) 
    return example      



from types import ModuleType
from thonnycontrib.backend.ast_parser import L1DocTestFlag
from thonnycontrib.backend.doctest_parser import ExampleWithExpected
from thonnycontrib.backend.evaluator import Evaluator
from thonnycontrib.exceptions import *
from thonnycontrib.backend.verdicts.ExceptionVerdict import ExceptionVerdict
from thonnycontrib.backend.verdicts.FailedVerdict import FailedVerdict
from thonnycontrib.backend.verdicts.PassedVerdict import PassedVerdict
from thonnycontrib.backend.doctest_parser import *
import unittest as ut
from thonnycontrib.tests.fixtures.backend_mock import *

# ########################################################
#    Tous les tests qui suivent utilisent l'invite `$$$`
#               AVEC une valeur attendue  
# ########################################################
class TestEvaluator(ut.TestCase):
    def setUp(self):
        self.evaluator = Evaluator(filename="<string>")
        self.mock_backend = backend_patch.start()
    
    def tearDown(self) -> None:
        del self.evaluator
        backend_patch.stop()
    
    """
    Ce test vérifie:
    1. Quand il n'y a aucun test dans une docstring alors aucun 
    type Example n'est produit. On assure plutôt que la liste
    d'Example associée au noeud est vide.
    2. Lorsqu'un l1doctest ne contient aucun test alors assurer
    que son flag est EMPTY_FLAG
    """
    def test_evaluate_when_empty_verdict(self):
        fake_source = \
"""
def f(a, b):
    if a < 0 and b < 0:
        return None
    return a + b
"""
        
        self.evaluator.set_source(fake_source)
        self.evaluator.set_module(self.__build_module_from_source(fake_source))
        
        # ###################################################
        # ------------- Vérification du verdict -------------
        l1doctests = self.evaluator.evaluate()
        
        # on assure qu'il existe un seul l1doctest
        self.assertTrue(len(l1doctests) == 1)
        l1doctest = l1doctests[0]
        self.assertTrue(l1doctest.is_evaluated()) # s'assure que le l1doctest a été évalué
        self.assertEqual(l1doctest.get_flag(), L1DocTestFlag.EMPTY_FLAG)
        
        # on assure qu'il n'y a aucun example
        self.assertTrue(len(l1doctest.get_examples()) == 0) 
    
    """
    Ce test vérifie:
    1. Le type `ExampleWithExpected` est le type `Example` extrait 
    par le doctest parser.
    2. Le verdict renvoyé est de type `PassedTest`. 
    """     
    def test_evaluate_when_passed_verdict(self):
        fake_source = \
"""
def f(a, b):
    '''
    $$$ f(1, 2)
    3
    '''
    if a < 0 and b < 0:
        return None
    return a + b
""" 
        self.evaluator.set_source(fake_source)
        self.evaluator.set_module(self.__build_module_from_source(fake_source))
        
        # ###########################################################
        # ------------- Vérification du type Example ----------------
        self._assert_example_type(ExampleWithExpected)
        
        # ###################################################
        # ------------- Vérification du verdict -------------
        l1doctests = self.evaluator.evaluate()
        
        # on assure qu'il existe un seul l1doctest avec ses examples
        self.assertTrue(len(l1doctests) == 1)
        l1doctest = l1doctests[0]
        self.assertTrue(l1doctest.is_evaluated()) # s'assure que le l1doctest a été évalué
        self.assertEqual(l1doctest.get_flag(), L1DocTestFlag.PASSED_FLAG)
        
        # on assure que le verdict de l'example est un passed verdict
        examples = l1doctest.get_examples()
        self.assertTrue(len(examples) == 1) # on assure qu'il existe un seul Example
        self.assertTrue(isinstance(l1doctest.get_verdict_from_example(examples[0]), PassedVerdict)) 
    
    
    """
    Ce test vérifie:
    1. Le type `ExampleWithExpected` est le type `Example` extrait 
    par le doctest parser.
    2. Quand un test échoue et que le résultat réél est un string vide,
    alors on assure que le message d'echec inclut une chaine viee. Ce test
    a été ajouté pour corriger le bug (issue #15, sur le lien suivant:
    https://gitlab.univ-lille.fr/mirabelle.nebut/thonny-tests/-/issues/15)
    """
    def test_evaluate_and_check_string_when_failed(self):
        fake_source = \
"""
def foo():
    '''
    $$$ foo()
    'a'
    '''
    return ''
""" 
        
        self.evaluator.set_source(fake_source)
        self.evaluator.set_module(self.__build_module_from_source(fake_source))
        
        # ###########################################################
        # ------------- Vérification du type Example ----------------
        self._assert_example_type(ExampleWithExpected)
        
        # ###################################################
        # ------------- Vérification du verdict -------------
        l1doctests = self.evaluator.evaluate()
        
        # on assure qu'il existe un seul l1doctest avec ses examples
        self.assertTrue(len(l1doctests) == 1)
        l1doctest = l1doctests[0]
        self.assertTrue(l1doctest.is_evaluated()) # s'assure que le l1doctest a été évalué
        self.assertEqual(l1doctest.get_flag(), L1DocTestFlag.FAILED_FLAG)
        
        examples = l1doctest.get_examples()
        self.assertTrue(len(examples) == 1) 
        verdict = l1doctest.get_verdict_from_example(examples[0])
        self.assertTrue(isinstance(verdict, FailedVerdict))
        self.assertTrue("Expected: 'a', Got: ''" in verdict.get_details())
    
    """
    Ce test vérifie:
    1. Le type `ExampleWithExpected` est le type `Example` extrait 
    par le doctest parser.
    2. Quand un test échoue et que le résultat réél est de type `str`,
    alors on assure que le message inclut bien les quotes. Ce test
    a été ajouté pour corriger le bug (issue #13, sur le lien suivant:
    https://gitlab.univ-lille.fr/mirabelle.nebut/thonny-tests/-/issues/13)
    """
    def test_evaluate_and_check_string_when_failed(self): 
        fake_source = \
"""
def toto():
    '''
    $$$ toto()
    "titi"
    '''
    return 'toto'
""" 
        
        self.evaluator.set_source(fake_source)
        self.evaluator.set_module(self.__build_module_from_source(fake_source))
        
        # ###########################################################
        # ------------- Vérification du type Example ----------------
        self._assert_example_type(ExampleWithExpected)
        
        # ###################################################
        # ------------- Vérification du verdict -------------
        l1doctests = self.evaluator.evaluate()
        
        # on assure qu'il existe un seul l1doctest avec ses examples
        self.assertTrue(len(l1doctests) == 1)
        l1doctest = l1doctests[0]
        self.assertTrue(l1doctest.is_evaluated()) # s'assure que le l1doctest a été évalué
        self.assertEqual(l1doctest.get_flag(), L1DocTestFlag.FAILED_FLAG)
        
        examples = l1doctest.get_examples()
        self.assertTrue(len(examples) == 1) 
        verdict = l1doctest.get_verdict_from_example(examples[0])
        self.assertTrue(isinstance(verdict, FailedVerdict))
        self.assertTrue("Expected: \"titi\", Got: 'toto'" in verdict.get_details())
    
    """
    Ce test vérifie:
    1. Le type `ExampleWithExpected` est le type `Example` extrait 
    par le doctest parser.
    2. Le verdict renvoyé est de type `FailedTest`. Enfin, on assure
    que le flag du l1doctest est FAILED_FLAG.
    """
    def test_evaluate_when_failed_verdict(self):
        fake_source = \
"""
def f(a, b):
    '''
    $$$ f(1, 2)
    2
    '''
    if a < 0 and b < 0:
        return None
    return a + b
""" 
        
        self.evaluator.set_source(fake_source)
        self.evaluator.set_module(self.__build_module_from_source(fake_source))
        
        # ###########################################################
        # ------------- Vérification du type Example ----------------
        self._assert_example_type(ExampleWithExpected)
        
        # ###################################################
        # ------------- Vérification du verdict -------------
        l1doctests = self.evaluator.evaluate()
        
        # on assure qu'il existe un seul l1doctest avec ses examples
        self.assertTrue(len(l1doctests) == 1)
        l1doctest = l1doctests[0]
        self.assertTrue(l1doctest.is_evaluated()) # s'assure que le l1doctest a été évalué
        self.assertEqual(l1doctest.get_flag(), L1DocTestFlag.FAILED_FLAG)
        
        examples = l1doctest.get_examples()
        self.assertTrue(len(examples) == 1) 
        verdict = l1doctest.get_verdict_from_example(examples[0])
        self.assertTrue(isinstance(verdict, FailedVerdict))
        self.assertTrue('Expected: 2, Got: 3' in verdict.get_details())
    
    """
    Ce test vérifie:
    1. Le type `ExampleWithExpected` est le type `Example` extrait 
    par le doctest parser.
    2. Le verdict renvoyé est de type `ExceptionTest` car la 
    comparaison `if a < 0 and b < 0:` va échouer si l'argument `a` ou `b`
    ne sont pas de type `int`. Enfin, on assure que le flag du l1doctest
    est FAILED_FLAG.
    """
    def test_evaluate_when_exception_verdict(self):
        fake_source = \
"""
def f(a, b):
    '''
    $$$ f('j', 2)
    2
    '''
    if a < 0 and b < 0:
        return None
    return a + b
""" 
        
        self.evaluator.set_source(fake_source)
        self.evaluator.set_module(self.__build_module_from_source(fake_source))
        
        # ###########################################################
        # ------------- Vérification du type Example ----------------
        self._assert_example_type(ExampleWithExpected)
        
        # ###################################################
        # ------------- Vérification du verdict -------------
        l1doctests = self.evaluator.evaluate()
        
        # on assure qu'il existe un seul l1doctest avec ses examples
        self.assertTrue(len(l1doctests) == 1)
        l1doctest = l1doctests[0]
        self.assertTrue(l1doctest.is_evaluated()) # s'assure que le l1doctest a été évalué
        self.assertEqual(l1doctest.get_flag(), L1DocTestFlag.FAILED_FLAG)
        
        examples = l1doctest.get_examples()
        self.assertTrue(len(examples) == 1) 
        verdict = l1doctest.get_verdict_from_example(examples[0])
        self.assertTrue(isinstance(verdict, ExceptionVerdict))
        self.assertTrue("TypeError: '<' not supported between instances of 'str' and 'int'"\
                            in verdict.get_details())
    
    """
    Ce test vérifie:
    1. Le type `ExampleWithExpected` est le type `Example` extrait 
    par le doctest parser.
    2. Quand la valeur attendue est None et que la fonction executée renvoie None
    alors on aura un verdict de type `PassedTest`. Le flag du l1doctest est 
    PASSED_FLAG.
    """    
    def test_evaluate_when_expected_is_none(self):
        fake_source = \
"""
def f(a, b):
    '''
    $$$ f(-1, -1)
    None
    '''
    if a < 0 and b < 0:
        return None
    return a + b
""" 
        
        self.evaluator.set_source(fake_source)
        self.evaluator.set_module(self.__build_module_from_source(fake_source))
        
        # ###########################################################
        # ------------- Vérification du type Example ----------------
        self._assert_example_type(ExampleWithExpected)
        
        # ###################################################
        # ------------- Vérification du verdict -------------
        l1doctests = self.evaluator.evaluate()
        
        # on assure qu'il existe un seul l1doctest avec ses examples
        self.assertTrue(len(l1doctests) == 1)
        l1doctest = l1doctests[0]
        self.assertTrue(l1doctest.is_evaluated()) # s'assure que le l1doctest a été évalué
        self.assertEqual(l1doctest.get_flag(), L1DocTestFlag.PASSED_FLAG)
        
        examples = l1doctest.get_examples()
        self.assertTrue(len(examples) == 1) 
        self.assertTrue(isinstance(l1doctest.get_verdict_from_example(examples[0]), PassedVerdict))
    
    """
    Ce test vérifie:
    1. Le type `ExampleWithExpected` est le type `Example` extrait 
    par le doctest parser.
    2. Quand il y a une erreur de syntax au niveau du test alors
    un verdict `ExceptionTest` est renvoyé. ENfin, on assure que le flag
    du l1doctest est FAILED_FLAG.
    """
    def test_evaluate_when_syntax_error(self):
        fake_source = \
"""
def f(a, b):
    '''
    $$$ f(1, 1
    2
    '''
    if a < 0 and b < 0:
        return None
    return a + b
""" 
        
        self.evaluator.set_source(fake_source)
        self.evaluator.set_module(self.__build_module_from_source(fake_source))
        
        # ###########################################################
        # ------------- Vérification du type Example ----------------
        self._assert_example_type(ExampleWithExpected)
        
        # ###################################################
        # ------------- Vérification du verdict -------------
        l1doctests = self.evaluator.evaluate()
        
        # on assure qu'il existe un seul l1doctest avec ses examples
        self.assertTrue(len(l1doctests) == 1)
        l1doctest = l1doctests[0]
        self.assertTrue(l1doctest.is_evaluated()) # s'assure que le l1doctest a été évalué
        self.assertEqual(l1doctest.get_flag(), L1DocTestFlag.FAILED_FLAG)
        
        examples = l1doctest.get_examples()
        self.assertTrue(len(examples) == 1) 
        verdict = l1doctest.get_verdict_from_example(examples[0])
        self.assertTrue(isinstance(verdict, ExceptionVerdict))
        self.assertTrue(SyntaxError.__name__ in verdict.get_details())
    
    """
    Ce test vérifie:
    1. Le type `ExampleWithExpected` est le type `Example` extrait 
    par le doctest parser.
    2. Quand la fonction executée utilise une variable non déclarée alors
    on assure que le verdict est de type `ExceptionTest`
    """
    def test_evaluate_when_runtime_error(self):
        fake_source = \
"""
def f(a, b):
    '''
    $$$ f(1, 1)
    2
    '''
    if a < 0 and b < 0:
        return None
    return a + b * c
""" 
        
        self.evaluator.set_source(fake_source)
        self.evaluator.set_module(self.__build_module_from_source(fake_source))
        
        # ###########################################################
        # ------------- Vérification du type Example ----------------
        self._assert_example_type(ExampleWithExpected)
        
        # ###################################################
        # ------------- Vérification du verdict -------------
        l1doctests = self.evaluator.evaluate()
        
        # on assure qu'il existe un seul l1doctest avec ses examples
        self.assertTrue(len(l1doctests) == 1)
        l1doctest = l1doctests[0]
        self.assertTrue(l1doctest.is_evaluated()) # s'assure que le l1doctest a été évalué
        self.assertEqual(l1doctest.get_flag(), L1DocTestFlag.FAILED_FLAG)
        
        examples = l1doctest.get_examples()
        self.assertTrue(len(examples) == 1) 
        verdict = l1doctest.get_verdict_from_example(examples[0])
        self.assertTrue(isinstance(verdict, ExceptionVerdict))
        self.assertTrue(NameError.__name__ in verdict.get_details())
        self.assertTrue("name 'c' is not defined" in verdict.get_details())  
    
    """
    Ce test vérifie:
    1. Le type `ExampleWithExpected` est le type `Example` extrait 
    par le doctest parser.
    2. Quand la valeur attendue est une exception et que la fonction lève 
    une exception alors on aura le verdict `ExceptionTest`.
    """      
    def test_evaluate_when_an_exception_is_expected_and_raised(self):
        fake_source = \
"""
def f(a, b):
    '''
    $$$ f(-1, -1)
    Exception
    '''
    if a < 0 and b < 0:
        raise Exception("a et b doivent être positifs")
    return a + b
""" 
        
        self.evaluator.set_source(fake_source)
        self.evaluator.set_module(self.__build_module_from_source(fake_source))
        
        # ###########################################################
        # ------------- Vérification du type Example ----------------
        self._assert_example_type(ExampleWithExpected)
        
        # ###################################################
        # ------------- Vérification du verdict -------------
        l1doctests = self.evaluator.evaluate()
        
        # on assure qu'il existe un seul l1doctest avec ses examples
        self.assertTrue(len(l1doctests) == 1)
        l1doctest = l1doctests[0]
        self.assertTrue(l1doctest.is_evaluated()) # s'assure que le l1doctest a été évalué
        self.assertEqual(l1doctest.get_flag(), L1DocTestFlag.FAILED_FLAG)
        
        examples = l1doctest.get_examples()
        self.assertTrue(len(examples) == 1) 
        verdict = l1doctest.get_verdict_from_example(examples[0])
        self.assertTrue(isinstance(verdict, ExceptionVerdict))
        self.assertTrue(Exception.__name__ in verdict.get_details())
        self.assertTrue("a et b doivent être positifs" in verdict.get_details()) 

    """
    Ce test vérifie:
    1. Le type `ExampleWithExpected` est le type `Example` extrait 
    par le doctest parser.
    2. Quand la valeur attendue est une exception mais la fonction ne lève pas
    une exception, dans ce cas on aura le verdict `FailedTest`.
    """        
    def test_evaluate_when_an_exception_is_expected_but_not_raised(self):
        fake_source = \
"""
def f(a, b):
    '''
    $$$ f(-1, -1)
    Exception
    '''
    if a < 0 and b < 0:
        return None
    return a + b
""" 
        
        self.evaluator.set_source(fake_source)
        self.evaluator.set_module(self.__build_module_from_source(fake_source))
        
        # ###########################################################
        # ------------- Vérification du type Example ----------------
        self._assert_example_type(ExampleWithExpected)
        
        # ###################################################
        # ------------- Vérification du verdict -------------
        l1doctests = self.evaluator.evaluate()
        
        # on assure qu'il existe un seul l1doctest avec ses examples
        self.assertTrue(len(l1doctests) == 1)
        l1doctest = l1doctests[0]
        self.assertTrue(l1doctest.is_evaluated()) # s'assure que le l1doctest a été évalué
        self.assertEqual(l1doctest.get_flag(), L1DocTestFlag.FAILED_FLAG)
        
        examples = l1doctest.get_examples()
        self.assertTrue(len(examples) == 1) 
        verdict = l1doctest.get_verdict_from_example(examples[0])
        self.assertTrue(isinstance(verdict, FailedVerdict))
        self.assertTrue(Exception.__name__ in verdict.get_details())
        self.assertTrue('Expected: Exception, Got: None' in verdict.get_details())
    
    """
    Ce test vérifie:
    1. Le type `ExampleWithoutExpected` est le type `Example` extrait 
    par le doctest parser.
    2. si un test est setup mais il existe quand même une valeur attendue
    alors le verdict renvoyée est de type `ExceptionTest`
    """
    def test_evaluate_when_setup(self):
        fake_source = \
"""
def f(a, b):
    '''
    $$$ a = 0
    0
    '''
    if a < 0 and b < 0:
        return None
    return a + b
""" 
        
        self.evaluator.set_source(fake_source)
        self.evaluator.set_module(self.__build_module_from_source(fake_source))
        
        # ###########################################################
        # ------------- Vérification du type Example ----------------
        self._assert_example_type(ExampleWithExpected)
        
        # ###################################################
        # ------------- Vérification du verdict -------------
        l1doctests = self.evaluator.evaluate()
        
        # on assure qu'il existe un seul l1doctest avec ses examples
        self.assertTrue(len(l1doctests) == 1)
        l1doctest = l1doctests[0]
        self.assertTrue(l1doctest.is_evaluated()) # s'assure que le l1doctest a été évalué
        self.assertEqual(l1doctest.get_flag(), L1DocTestFlag.FAILED_FLAG)
        
        examples = l1doctest.get_examples()
        self.assertTrue(len(examples) == 1) 
        verdict = l1doctest.get_verdict_from_example(examples[0])
        self.assertTrue(isinstance(verdict, ExceptionVerdict))
        self.assertTrue(SyntaxError.__name__ in verdict.get_details())
        self.assertTrue("invalid syntax. Maybe you meant '==' or ':=' instead of '='?"\
                            in verdict.get_details())
    
    
    def _assert_example_type(self, example_type: Example):
        """Assert that the expected `Example` type is of type the given `example_type`"""
        l1doctests = self.evaluator.get_test_finder().find_l1doctests()
        
        # on assure qu'il existe un seul l1doctest
        self.assertTrue(len(l1doctests) == 1)
        for l1_docTest in l1doctests:
            examples = l1_docTest.get_examples()
            self.assertTrue(len(examples) == 1) # on assure qu'il existe un seul type Example 
            self.assertTrue(isinstance(examples[0], example_type))
    
    def __build_module_from_source(self, source: str) -> ModuleType:
        """
        Build a module containing the functions declared in the given `source`.
        """
        from types import ModuleType
        fake_module = ModuleType(self.evaluator.get_filename())
        exec(source, fake_module.__dict__)
        return fake_module


if __name__ == '__main__':
    ut.main(verbosity=2)   
        



from thonnycontrib.l1test_frontend.l1test_runner import L1TestRunner
from thonnycontrib.l1test_frontend.l1test_reporter import L1TestTreeView, L1TestErrorView
from thonnycontrib.backend.doctest_parser import *
from unittest.mock import *
from thonny.workbench import WorkbenchEvent
import unittest as ut, thonny
from thonnycontrib.properties import PLUGIN_NAME
from thonnycontrib.tests.fixtures.workbench_mock import *


# ##############################################################
#    Tous les tests qui suivent testent la classe L1TestRunner
# ##############################################################


class TestTestRunner(ut.TestCase):
    def setUp(self) -> None:
        self.workbench_event = WorkbenchEvent(sequence="BackendRestart")
    
    def tearDown(self) -> None:
        del self.workbench_event 
    
    """
    Quand un restart backend complet(ex.via le button rouge) est réalisé,
    alors on doit s'assurer que:
    - L'état de `l1TestRunner` est déclaré comme terminé.
    - Le contenu de la treeview est vide.
    """
    @patch("thonny.get_workbench", return_value=MockWorkbench())
    def test_restart_backend_event_when_complete_restart(self, mock_get_workbench: MagicMock):
        self.__add_views()
        l1test_runner = L1TestRunner()
        
        # on s'assure que le mock_get_workbench est appelée lors de l'instantiation de L1TestRunner
        mock_get_workbench.assert_called()
        
        # On doit ajouter du faux contenus à la treeview, pour vérifier qu'elle
        # sera vraiment nettoyée après le restart_backend.
        l1test_treeview = l1test_runner.get_reporter().get_treeview()
        treeview = self.__add_fake_row_in_treeview(l1test_treeview)
        
        self.workbench_event.full = True # on suppose que c'est un redémarrage complet
        l1test_runner._on_restart_backend(self.workbench_event)

        # on s'assure qu'après le redémarrage complet du backend l'état du `L1TestRunner` est déclaré 
        # comme terminé.
        self.assertFalse(l1test_runner.is_running())
        
        # on s'assure que le contenu de la treeview est vide
        self.assertTrue(len(treeview.get_children()) == 0)
    
    """
    Si un restart backend partiel(ex. invocation d'une nouvelle commande magique) est appelé
    après l'invocation de la commande magique `L1test`, alors on doit s'assurer que:
    - Le `l1TestRunner` est en execution.
    - La treeview et effacée.
    
    Le cas de ce test : la treeview était initialement vide.
    """
    @patch("thonny.get_workbench", return_value=MockWorkbench())
    def test_restart_backend_event_when_partial_restart_and_when_l1test_is_invoked_case1(self,  
        mock_get_workbench: MagicMock
    ):
        
        self.__add_views()
        l1test_runner = L1TestRunner()
        
        # on s'assure que le mock_get_workbench est appelée lors de l'instantiation de L1TestRunner
        mock_get_workbench.assert_called()
        
        # on suppose que la commande L1test a été invoquée
        l1test_runner.set_is_running()
        # On s'assure que le `L1TestRunner` est en execution.
        self.assertTrue(l1test_runner.is_running())
        
        self.workbench_event.full = False # on suppose que c'est un redémarrage partiel
        l1test_runner._on_restart_backend(self.workbench_event)

        treeview = l1test_runner.get_reporter().get_treeview().get_treeview()
        # on s'assure qu'il existe rien dans la treeview
        self.assertTrue(len(treeview.get_children()) == 0)

    """
    Si un restart backend partiel(ex. invocation d'une nouvelle commande magique) est appelé
    après l'invocation de la commande magique `L1test` alors on doit s'assurer que:
    - Le `l1TestRunner` est en execution.
    - La treeview est vide.
    
    Le cas de ce test : la treeview était initialement remplie.
    """
    @patch("thonny.get_workbench", return_value=MockWorkbench())
    def test_restart_backend_event_when_partial_restart_and_when_l1test_is_invoked_case2(self,  
        mock_get_workbench: MagicMock
    ):
        self.__add_views()
        l1test_runner = L1TestRunner()
        
        # on s'assure que le mock_get_workbench est appelée lors de l'instantiation de L1TestRunner
        mock_get_workbench.assert_called()
        
        # on suppose que la commande L1test a été invoquée
        l1test_runner.set_is_running()
        # On s'assure que le `L1TestRunner` est en execution.
        self.assertTrue(l1test_runner.is_running())
        
        # On doit ajouter du faux contenus à la treeview, pour vérifier que son contenue
        # sera effacée et que le message "Executing tests in progress ..." apparaîtera à la place
        l1test_treeview = l1test_runner.get_reporter().get_treeview()
        treeview = self.__add_fake_row_in_treeview(l1test_treeview)
        
        self.workbench_event.full = False # on suppose que c'est un redémarrage partiel
        l1test_runner._on_restart_backend(self.workbench_event)
        
        # On s'assure qu'il y a rien dans la treeview
        self.assertTrue(len(treeview.get_children()) == 0)
    
    """
    Si un restart backend partiel(ex. invocation d'une nouvelle commande magique) est appelé
    après une autre commande magique(pas `L1test`), alors on doit s'assurer que:
    - Le `l1TestRunner` n'est pas en execution.
    - Le contenu de la treeview n'est pas effacée.
    """
    @patch("thonny.get_workbench", return_value=MockWorkbench())
    def test_restart_backend_event_when_partial_restart_and_when_l1test_is_not_invoked(self,  
            mock_get_workbench: MagicMock
    ):
        self.__add_views()
        l1test_runner = L1TestRunner()
        
        # on s'assure que le mock_get_workbench est appelée lors de l'instantiation de L1TestRunner
        mock_get_workbench.assert_called()
        
        # On s'assure que le `L1TestRunner` n'est pas en execution.
        self.assertFalse(l1test_runner.is_running())
        
        # On doit ajouter du faux contenus à la treeview, pour vérifier que le contenu
        # de la treeview ne sera pas du tout effacée.
        l1test_treeview = l1test_runner.get_reporter().get_treeview()
        row_text = "A fake content"
        treeview = self.__add_fake_row_in_treeview(l1test_treeview, row_text=row_text)
        
        self.workbench_event.full = False # on suppose que c'est un redémarrage partiel
        l1test_runner._on_restart_backend(self.workbench_event)
        
        # on s'assure que le contenu de la treeview reste le même 
        self.assertTrue(len(treeview.get_children()) == 1)
        for child_id in treeview.get_children():
            # on doit vérifier le contenu de la ligne ajoutée
            self.assertTrue(row_text in treeview.item(child_id)["text"])
    
    """
    Adds one fake row to the treeview and asserts that the row is well
    added to the treeview.
    
    Returns: 
        Widget(tk.Treeview): if the assertions are passed so the treeview with the added row
        will be returned.
    """
    def __add_fake_row_in_treeview(self, l1test_treeview: L1TestTreeView, row_text="A fake content"):
        # on s'assure que la ligne est bien ajoutée à la treeview
        treeview = l1test_treeview.get_treeview()
        treeview.insert("", "end", text=row_text)
        self.assertTrue(len(treeview.get_children()) == 1)
        for child_id in treeview.get_children():
            # on doit vérifier le contenu de la ligne ajoutée
            self.assertTrue(row_text in treeview.item(child_id)["text"])
        return treeview
        
    def __add_views(self):
        Workbench = thonny.get_workbench()
        Workbench.add_view(L1TestTreeView, PLUGIN_NAME, "nw")
        Workbench.add_view(L1TestErrorView, '%s errors' % PLUGIN_NAME, "sw")
        

if __name__ == '__main__':
    ut.main(verbosity=2)   



from tkinter import ttk
import tkinter as tk
from thonny import rst_utils, tktextext, ui_utils
import thonny, tkinter.font as tk_font
from thonny.ui_utils import scrollbar_style
from thonnycontrib.utils import get_font_size_option, get_font_family_option

class L1TestErrorView(tktextext.TextFrame):
    def __init__(self, master=None):
        self.text_frame: tktextext.TextFrame = super().__init__(
            master,
            text_class=AssistantRstText,
            vertical_scrollbar_style=scrollbar_style("Vertical"),
            horizontal_scrollbar_style=scrollbar_style("Horizontal"),
            horizontal_scrollbar_class=ui_utils.AutoScrollbar,
            read_only=True,
            wrap="word",
            font="TkDefaultFont",
        )
        self.workbench = thonny.get_workbench()

        self.title_font = tk_font.Font(family=get_font_family_option(),
                                  size=get_font_size_option(),
                                  weight="bold")
        
        self.text.tag_config("green", foreground="darkgreen")
        self.text.tag_config("red", foreground="red")
        self.text.tag_config("darkred", foreground="#d41919")
        self.text.tag_config("yellow", foreground="yellow")  
        
        self.text.tag_configure("title", font=self.title_font)              
    
    def append_text(self, chars, tags=()):
        self.text.direct_insert("end", chars, tags=tags)
        self.update_font(font_family_opt="io_font_family")

    def write(self, data):
        self.text.set_content(data)

    def append_rst(self, rst, tags=[]):
        tags += ["url"]
        self.text.append_rst(rst, tags)
        self.update_font(font_family_opt="io_font_family")
        
    def update_font(self, font_family_opt=""):
        """Met à jour la police et la taille de la police du texte.

        Args:
            font_family_opt (str, optional): L'option de la famille de police. Defaults to "".
        """
        self.text.tag_configure("title", font=self.title_font)   
        self.text.config(font=(get_font_family_option(option=font_family_opt), 
                               get_font_size_option()))

    def clear(self):
        self.write("")
        
    def hide_view(self):
        self.workbench.hide_view(self.__class__.__name__)
    
    def show_view(self):
        self.workbench.show_view(self.__class__.__name__)
        
class AssistantRstText(rst_utils.RstText):
    def configure_tags(self):
        super().configure_tags()
        self.workbench = thonny.get_workbench()
        font = tk_font.Font(
                    family=get_font_family_option(option="io_font_family"),
                    size=get_font_size_option(),
                    underline=True
                )
        
        self.tag_configure("url", font=font)

        self.tag_raise("sel")


r""" 
Içi, vous retrouverez les variables d'environnements utilisées par le l1test_backend
-----------------------------------------------------------------------------------

Il n'existe pas de moyens pour acheminer les données necessaires au traitement 
de la commande par l1test_backend. Pour cela, les plugins backend existant passent 
par enregistrer les données comme variable d'environement, ainsi le l1test_backend 
peut retrouver la donnée et décider comment executer ses tests. 

Note: Les variables d'environnements ne persistent pas après la fermeture de Thonny.
"""

# `IMPORT_MODULE_VAR` est une variable d'environnment qui stocke la valeur de l'option
# `l1test_options.IMPORT_MODULE` qui indique si oui(`True`) ou non(`False`) qu'il faut
# importer le module executé dans le shell.
IMPORT_MODULE_VAR = "import_module"

# `SELECTED_LINE_VAR` est une variable d'environnment qui stocke le numéro de ligne de 
# la fonction séléctionnée si `IS_SELECTED_VAR` est à True, sinon elle stocke `None`.
SELECTED_LINE_VAR = "selected_line"



from thonny.workbench import Workbench
from typing import Any, Callable
import tkinter as tk

class MockWorkbench(Workbench):
    def __init__(self):
        self._view_records = {}
        pass
    
    def add_view(self, cls, label: str, default_location: str, visible_by_default: bool = False, default_position_key = None,
    ) -> None:
        view_id = cls.__name__
        self._view_records[view_id] = {
                "class": cls,
                "label": label,
                "location": "nw",
                "position_key": "A",
                "visibility_flag": None,
            }
        
    def get_view(self, view_id: str, create: bool = True) -> tk.Widget:
        if "instance" not in self._view_records[view_id]:
            if not create:
                raise RuntimeError("View %s not created" % view_id)
            class_ = self._view_records[view_id]["class"]
            # create the view
            view = class_()  # View's master is workbench to allow making it maximized
            self._view_records[view_id]["instance"] = view
            view.hidden = True
        return self._view_records[view_id]["instance"]
    
    def get_option(self, name: str, default=None) -> Any:
        if name in ("view.editor_font_family", "view.io_font_family"):
            return ""
        elif  name == "view.editor_font_size":
            return 0
        # Need to return Any, otherwise each typed call site needs to cast
        return self._configuration_manager.get_option(name, default)
    
    def bind(self, sequence: str, func: Callable, add: bool = None) -> None:
        pass
    
    def show_view(self, view_id: str, set_focus: bool = True) :
        return 
    
    def hide_view(self, view_id: str) :
        return None



import re
import sys
import textwrap
from typing import List

from .doc_template import DocTemplate
from ..backend.ast_parser import L1TestAstParser
from .doc_template import *
from ..properties import CANNOT_GENERATE_THE_DOCSTRING
from thonny.tktextext import *  
from ..exceptions import DocGeneratorParserException, FrontendException, NoFunctionSelectedToDocumentException
from ..utils import replace_error_line, get_last_exception
from thonny.editors import EditorCodeViewText
from ..ThonnyLogsGenerator import log_doc_in_thonny
from thonny import get_workbench
from .. l1test_frontend import get_l1test_runner
from thonnycontrib import docstring_generator

r""" Docstring Generator Module
Description:
------------
This module generates a docstring using the templates.

For a selected line the `DocGenerator` tries to verify if the selected line
corresponds to a function signature. If the selected line is a function 
signature so a the `DocGenerator` will build a custom function with the given
selected line and then it parses(with AST parser) this function. If the AST parser
fails so the error will be displayed in the ErrorView. otherwise, the docstring
will be generated.

About templates, the docstring generator invokes the `DefaultDocTemplate` by default. 
The `DocTemplate.DefaultDocTemplate` class contains an implementation 
of a default template. 

How to use the Generator in thonny IDE:
---------------------------------------
- Right click on a function or a class declaration(it's prototype) and choose 
in the `edit menu` ~Generate Docstring~ button. You can also select the short cut 
Alt+d after putting the cursor on the function(or a class) declaration.

- Just a return on a function declaration will generate its docstring.
"""

class DocParser:
    def __init__(self, filename="", source=""):
        self._filename = filename
        self._source = source
    
    def parse_source(self, error_line:int=None):
        """
        Parses the given `source` and returns a list of 
        AST nodes that may contains a doctsring.
        
        As the AST python module stated in its documentation, the AST nodes that can 
        contain a dosctring are reported at ~SourceParser.SUPPORTED_TYPES~.
        
        Args:
            error_line(int): Set it only if you want to change the error line in 
            the raised exception. If `None`, so the error line mentioned in the error will 
            be kept. 
        
        Raises:
            DocGeneratorParserException: if the ast parser is failed.
        """
        try :
            parsed = ast.parse(self._source, self._filename, mode="single").body
            return parsed[0] if len(parsed) == 1 else None
        except Exception as e: # if a compilation error occurs during the parsing
            error_info = sys.exc_info()
            last_exception = get_last_exception(error_info)
            if error_line:
                last_exception = replace_error_line(last_exception, error_line)
            raise DocGeneratorParserException(last_exception)    
    
    def get_filename(self):
        return self._filename
    
    def set_filename(self, filename: str):
        self._filename = filename
    
    def set_source(self, source: str):
        self._source = source
        
    def get_source(self):
        return self._source 
    
    def set_ast_parser(self, parser) :
        self._ast_parser = parser

class DocGenerationStrategy(ABC):
    _SIGNATURE_REGEX = r"\s*(?P<id>def|class)\s*.*\s*$"
    
    def __init__(self, text_widget:EditorCodeViewText, parser:DocParser=DocParser()):
        self._parser = parser 
        self._text_widget = text_widget
    
    @abstractmethod
    def can_generate(self, selected_lineno:int) -> bool:
        pass
    
    @abstractmethod
    def generate(self) -> None:
        pass
    
    def _generate(self, signature:str, selected_lineno:int=None) -> str:   
        """Generate a docstring from a given signature (or prototype).
        
        Args:
            - signature(str): The signature for which the docstring will be generated.
            The signature should always be finished by a ":" caractere, otherwise the 
            docstring not will be generated. 
            - selected_lineno(int, Optional): This parameter is optional. It is the line number
            of the signature. This will be usefull for errors raised by the AST parser.
            If the ast parser raises an exception, the line number of the exception will be set to
            the given lineno.
            - text_widget(EditorCodeViewText, Optional): The view in which the generated docstring will 
            be inserted. Set to `None` if you want just to get the generated docstring. If `None` the
            generated docstring will not be inserted in any widget.
        
        Return:
            str: returns the generated docstring.
        
        Raises:
            - NoFunctionSelectedToDocument: When a selected line don't corresponds to 
            a function declaration. 
            - DocGeneratorParserException: when the ast parser fails.
        """
        if signature is None: signature = ""
        
        # We should check that the line is a function declaration and that ends with ':' character. 
        declaration_match = re.match(self._SIGNATURE_REGEX, signature)

        if not declaration_match:
            raise NoFunctionSelectedToDocumentException("No signature is selected to document!\n")
        else:   
            id_signature = declaration_match.group("id") # c'est le tag <id> dans l'expression régulière
            
            template:DocTemplate = DocTemplateFactory.create_template(id_signature) 
            
            generated_temp = self.__get_generated_template(template, signature, selected_lineno)
            
            indent = self.__compute_indent(signature)
            generated_doc = textwrap.indent(generated_temp, indent)
            if self._text_widget:
                # c'est içi que la docstring est ajoutée à l'éditeur
                self._text_widget.insert(str(selected_lineno + 1) + ".0", generated_doc)
            return generated_doc
    
    def __get_generated_template(self, template:DocTemplate, signature:str, selected_lineno:int) -> str:
        """
        Creates a custom function with the given `signature` then parses this function and if the
        AST parser success so the docstring will be generated. If the AST parser fails, so the 
        reported exception will be raised.

        Args:
            signature (str): the signature for which the docstring will be generated.
            selected_lineno (int): the line that corresponds to the selected line. This arg is used 
            to change the error line in the reported exception to the given `selected_line`. Remember
            that the error line will be always "1" if the ast parser fails, because the parsed source
            contains only the custom function.
            
        Returns:
            (str): The generated template.
            
        Raises: 
            DocGeneratorParserException: when the ast parser fails.
        """
        # The approach is to take the signature of the selected function then
        # adds a custom body to this function.
        custom_func = self._create_custom_body(signature)
        
        # don't forgot that the result of parsing is a list of supported nodes
        # -> see the doc
        self._parser.set_source(custom_func)
        node = self._parser.parse_source(selected_lineno)
        
        # Generate an event in Thonny with l1test/ThonnyLogsGenerator.log_doc_in_thonny
        log_doc_in_thonny(node)
        
        return template.get_template(node)

    def __compute_indent(self, signature:str) -> int:
        """
        Get the indentation based on the whitespaces located in the given `signature`.

        Args:
            signature (str): a signature of a function

        Returns:
            int: returns the indentation based on the whitespaces located in the given `signature`.
        """
        space_match = re.search("^(\s+)", signature)
        python_indent = 4
        sig_indent = len(space_match.group(1)) if space_match else 0
        return " " * (sig_indent + python_indent)
    
    def _create_custom_body(self, signature:str):
        signature = signature.strip()
        # on supprime tous ce qu'il vient après les ":" dans la signature. Car, on veut juste la signature.
        # En python on peut avoir une syntaxe comme : def foo(): pass, cela pose problème pour l'ast parser
        # il faut donc qu'on ait juste la signature sans le corps de la fonction.
        signature = re.sub(r"\s*(?P<id>def|class)(?P<signature>\s*.*\s*:)(?P<body>\s*.*)$", r"\g<id>\g<signature>", signature)
        indent = " " * 4
        return signature + "\n" + indent + "pass"
    
    def set_parser(self, parser: DocParser):
        self._parser = parser
        
    def get_parser(self):
        return self._parser
    
    def set_filename(self, filename):
        self._parser.set_filename(filename)
        
    def set_text_widget(self, text_widget:EditorCodeViewText):
        self._text_widget = text_widget
    
class AutoGenerationStrategy(DocGenerationStrategy):
    def __init__(self, text_widget:EditorCodeViewText=None, parser=DocParser()):
        super().__init__(text_widget, parser)
        self.__selected_sig = None
        self.__selected_lineno = None
    
    def can_generate(self, selected_lineno:int) -> bool:
        selected_sig = self._text_widget.get(str(selected_lineno)+".0", str(selected_lineno+1)+".0").strip().strip("\n")
        if selected_sig != "":
            self.__selected_sig = selected_sig
            self.__selected_lineno = selected_lineno
        return self.__selected_sig != ""
    
    def generate(self):
        return super()._generate(self.__selected_sig, self.__selected_lineno)
        
class ManualGenerationStrategy(DocGenerationStrategy):
    _OUTLINER_REGEX = r"\s*(?P<type>def|class)[ ]+(?P<name>[\w]+)"
    
    def __init__(self, text_widget:EditorCodeViewText=None, parser=DocParser()):
        super().__init__(text_widget, parser)
        self.__nodes: List[SourceNode] = []
        self.__selected_node = None
    
    def can_generate(self, selected_lineno:int) -> bool:
        self.__nodes = self.__parse(self._text_widget.get("1.0", "end"))
        self.__selected_node = next((node for node in self.__nodes if node.get_starting_lineno() <= selected_lineno <= node.get_ending_lineno()), None)
        return self.__selected_node is not None
    
    def generate(self) -> str:
        return super()._generate(self.__selected_node.get_signature(), self.__selected_node.get_starting_lineno())
     
    def __parse(self, source:str) :
        """
        Parses a source and returns a list of the outlined nodes. 
        The outlined nodes are either a class or a function. For 
        each outlined node an object of type `OutlinedNode` is built 
        in which we store the type (class/function), the name and the lineno
        of the outlined node.
        """
        current_node = None
        lines = source.splitlines()
        lineno = 1
        last_non_empty_line = None
        for line in lines:
            match = re.match(self._OUTLINER_REGEX, line)
            if match:
                # If there was a previous outlined node, set its ending lineno.
                if current_node:
                    current_node.set_ending_lineno(last_non_empty_line or current_node.get_starting_lineno())
                    yield current_node

                # Create a new outlined node.
                current_node = SourceNode(line, lineno, None)
                last_non_empty_line = None  # Reset the last_non_empty_line for the new node.

            elif line.strip():
                last_non_empty_line = lineno  # Update last_non_empty_line.

            lineno += 1

        # Set the ending lineno for the last outlined node.
        if current_node:
            current_node.set_ending_lineno(last_non_empty_line or current_node.get_starting_lineno())
            yield current_node
    
    def get_nodes(self):
        return self.__nodes
        
class SourceNode():
    """
    This class represents an outlined node. An outlined node is either a class or a function.
    """ 
    def __init__(self, signature:str, starting_lineno:int, ending_lineno) -> None:
        self.__signature = signature
        self.__starting_lineno = starting_lineno
        self.__ending_lineno = ending_lineno
        
    def get_starting_lineno(self):
        return self.__starting_lineno  
    
    def get_ending_lineno(self):
        return self.__ending_lineno
    
    def set_ending_lineno(self, ending_lineno):
        self.__ending_lineno = ending_lineno
        
    def set_signature(self, signature):
        self.__signature = signature
    
    def get_signature(self):
        return self.__signature
    
    def __str__(self) -> str:
        return f"signature: {self.__signature}, s_lineno: {self.__starting_lineno}, e_lineno: {self.__ending_lineno}" 
    
    
class DocGenerator():
    def __init__(self, strategy:DocGenerationStrategy=AutoGenerationStrategy()):
        docstring_generator._doc_generator = self
        self._has_exception = False 
        self._strategy = strategy
   
    def run(self, selected_lineno:int, text_widget:EditorCodeViewText):
        try:
            self.__run(selected_lineno, text_widget)
        except NoFunctionSelectedToDocumentException as e:
            pass # do nothing
        except FrontendException as e: # parsing error
            self.set_has_exception(True)
            self._show_error(str(e))

    def __run(self, selected_lineno:int, text_widget:EditorCodeViewText):
        self._strategy.set_text_widget(text_widget)
        if self._strategy.can_generate(selected_lineno):
            filename = get_workbench().get_editor_notebook().get_current_editor().get_filename()
            if not filename:
                filename = "<unknown>" 
                    
            self._strategy.set_filename(filename)
            self._strategy.generate()   
            
            # après la génération (réussie) on vérifie si docgen avait rencontré une exception avant. Si oui, 
            # on supprime l'exception de docgen (car elle a été déja montrée).
            if get_l1test_runner().has_exception() or self.has_exception(): # si docgen avait lancé une exception avant
                # si les deux ont lancé une exception, on affiche l'exception de docgen
                if get_l1test_runner().has_exception() and self.has_exception(): 
                    get_l1test_runner().clean_error_view()
                    get_l1test_runner().get_reporter().get_error_view().hide_view()
                    get_l1test_runner().get_reporter().get_treeview().show_view()
                else:
                    self._show_treeview()
                
                self.set_has_exception(False) # success
    
    def _show_treeview(self):
        """
        Cleans the ErrorView and hides it. Retreives the Treeview and shows it.
        """
        get_l1test_runner().hide_errorview_and_show_treeview()
    
    def _show_error(self, error_msg:str, error_title:str=CANNOT_GENERATE_THE_DOCSTRING):
        """
        Shows the error in the ErrorView if the docstring generator raises an exception.
        """
        l1test_runner = get_l1test_runner()
        if self.has_exception():
            l1test_runner.show_errors(exception_msg=error_msg, title=error_title)
            l1test_runner.get_reporter().get_error_view().show_view() 
            l1test_runner.get_reporter().get_treeview().hide_view()     
        
    def has_exception(self):
        return self._has_exception
    
    def set_has_exception(self, has_exception: bool):
        self._has_exception = has_exception
        
    def get_strategy(self):
        return self._strategy
    
    def set_strategy(self, strategy: DocGenerationStrategy):
        self._strategy = strategy


# Auteur : Esteban COLLARD, Nordine EL AMMARI
# Refactored by Réda ID-TALEB

from .ExampleVerdict import ExampleVerdict

class FailedVerdict(ExampleVerdict):
    def __init__(self, filename:str, lineno, tested_line, expected_result, real_result):
        super().__init__(filename, lineno, tested_line, expected_result, real_result)
 
    def get_color(self):
        return "red"

    def isSuccess(self):
        return False
    
    def get_details(self):
        want, got = self.expected_result, self.details    
        if isinstance(got, str): # "''"
            got = "''" if got == str() else "'%s'" % got
        return 'Expected: %s, Got: %s' % (want, got)
    
    def _verdict_message(self):
        return "Test failed"